import base64
import logging
from unittest import mock, TestCase

from pcs_test.tools.assertions import (
    assert_raise_library_error,
    assert_report_item_list_equal,
)
from pcs_test.tools.custom_mock import MockLibraryReportProcessor

from pcs.common import reports
from pcs.common.reports import ReportItemSeverity as severity
from pcs.common.services.interfaces import ServiceManagerInterface
from pcs.common.services.errors import (
    DisableServiceError,
    EnableServiceError,
    StartServiceError,
    StopServiceError,
)
from pcs.common.reports import codes as report_codes
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.external import KillServicesError

import pcs.lib.commands.qdevice as lib


class QdeviceTestCase(TestCase):
    def setUp(self):
        self.mock_logger = mock.MagicMock(logging.Logger)
        self.mock_reporter = MockLibraryReportProcessor()
        mock.patch(
            "pcs.lib.env.get_service_manager", spec=ServiceManagerInterface
        ).start()
        self.addCleanup(mock.patch.stopall)
        self.lib_env = LibraryEnvironment(self.mock_logger, self.mock_reporter)


class QdeviceBadModelTest(QdeviceTestCase):
    @staticmethod
    def base_test(func):
        assert_raise_library_error(
            func,
            (
                severity.ERROR,
                report_codes.INVALID_OPTION_VALUE,
                {
                    "option_name": "model",
                    "option_value": "bad model",
                    "allowed_values": ["net"],
                    "cannot_be_empty": False,
                    "forbidden_characters": None,
                },
            ),
        )

    def test_setup(self):
        self.base_test(
            lambda: lib.qdevice_setup(self.lib_env, "bad model", False, False)
        )

    def test_destroy(self):
        self.base_test(lambda: lib.qdevice_destroy(self.lib_env, "bad model"))

    def test_status_text(self):
        self.base_test(
            lambda: lib.qdevice_status_text(self.lib_env, "bad model")
        )

    def test_enable(self):
        self.base_test(lambda: lib.qdevice_enable(self.lib_env, "bad model"))

    def test_disable(self):
        self.base_test(lambda: lib.qdevice_disable(self.lib_env, "bad model"))

    def test_start(self):
        self.base_test(lambda: lib.qdevice_start(self.lib_env, "bad model"))

    def test_stop(self):
        self.base_test(lambda: lib.qdevice_stop(self.lib_env, "bad model"))

    def test_kill(self):
        self.base_test(lambda: lib.qdevice_kill(self.lib_env, "bad model"))


@mock.patch("pcs.lib.commands.qdevice.qdevice_net.qdevice_setup")
@mock.patch.object(LibraryEnvironment, "cmd_runner", lambda self: "mock_runner")
class QdeviceNetSetupTest(QdeviceTestCase):
    # pylint: disable=no-member
    def test_success(self, mock_net_setup):
        lib.qdevice_setup(self.lib_env, "net", False, False)

        mock_net_setup.assert_called_once_with("mock_runner")
        self.assertEqual(self.lib_env.service_manager.method_calls, [])
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    report_codes.QDEVICE_INITIALIZATION_SUCCESS,
                    {
                        "model": "net",
                    },
                )
            ],
        )

    def test_start_enable_success(self, mock_net_setup):
        lib.qdevice_setup(self.lib_env, "net", True, True)

        mock_net_setup.assert_called_once_with("mock_runner")
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [
                mock.call.enable("corosync-qnetd"),
                mock.call.start("corosync-qnetd"),
            ],
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    report_codes.QDEVICE_INITIALIZATION_SUCCESS,
                    {
                        "model": "net",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_ENABLE,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_START,
                        "service": "quorum device",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_START,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
            ],
        )

    def test_init_failed(self, mock_net_setup):
        mock_net_setup.side_effect = LibraryError("mock_report_item")
        self.assertRaises(
            LibraryError,
            lambda: lib.qdevice_setup(self.lib_env, "net", False, False),
        )
        mock_net_setup.assert_called_once_with("mock_runner")
        self.assertEqual(self.lib_env.service_manager.method_calls, [])
        assert_report_item_list_equal(self.mock_reporter.report_item_list, [])

    def test_enable_failed(self, mock_net_setup):
        self.lib_env.service_manager.enable.side_effect = EnableServiceError(
            "test service", "test error"
        )

        assert_raise_library_error(
            lambda: lib.qdevice_setup(self.lib_env, "net", True, True),
            (
                severity.ERROR,
                reports.codes.SERVICE_ACTION_FAILED,
                {
                    "action": reports.const.SERVICE_ACTION_ENABLE,
                    "service": "test service",
                    "reason": "test error",
                    "node": "",
                    "instance": "",
                },
            ),
        )

        mock_net_setup.assert_called_once_with("mock_runner")
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.enable("corosync-qnetd")],
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    report_codes.QDEVICE_INITIALIZATION_SUCCESS,
                    {
                        "model": "net",
                    },
                )
            ],
        )

    def test_start_failed(self, mock_net_setup):
        self.lib_env.service_manager.start.side_effect = StartServiceError(
            "test service", "test error"
        )

        assert_raise_library_error(
            lambda: lib.qdevice_setup(self.lib_env, "net", True, True),
            (
                severity.ERROR,
                reports.codes.SERVICE_ACTION_FAILED,
                {
                    "action": reports.const.SERVICE_ACTION_START,
                    "service": "test service",
                    "reason": "test error",
                    "node": "",
                    "instance": "",
                },
            ),
        )

        mock_net_setup.assert_called_once_with("mock_runner")
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [
                mock.call.enable("corosync-qnetd"),
                mock.call.start("corosync-qnetd"),
            ],
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    report_codes.QDEVICE_INITIALIZATION_SUCCESS,
                    {
                        "model": "net",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_ENABLE,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_START,
                        "service": "quorum device",
                        "instance": "",
                    },
                ),
            ],
        )


@mock.patch("pcs.lib.corosync.qdevice_net.qdevice_status_cluster_text")
@mock.patch("pcs.lib.commands.qdevice.qdevice_net.qdevice_destroy")
@mock.patch.object(LibraryEnvironment, "cmd_runner", lambda self: "mock_runner")
class QdeviceNetDestroyTest(QdeviceTestCase):
    # pylint: disable=no-member
    def test_success_not_used(self, mock_net_destroy, mock_status):
        mock_status.return_value = ""

        lib.qdevice_destroy(self.lib_env, "net")

        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [
                mock.call.stop("corosync-qnetd"),
                mock.call.disable("corosync-qnetd"),
            ],
        )
        mock_net_destroy.assert_called_once_with()
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_DISABLE,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    report_codes.QDEVICE_DESTROY_SUCCESS,
                    {
                        "model": "net",
                    },
                ),
            ],
        )

    def test_success_used_forced(self, mock_net_destroy, mock_status):
        mock_status.return_value = 'Cluster "a_cluster":\n'

        lib.qdevice_destroy(self.lib_env, "net", proceed_if_used=True)

        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [
                mock.call.stop("corosync-qnetd"),
                mock.call.disable("corosync-qnetd"),
            ],
        )
        mock_net_destroy.assert_called_once_with()
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.WARNING,
                    report_codes.QDEVICE_USED_BY_CLUSTERS,
                    {
                        "clusters": ["a_cluster"],
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_DISABLE,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    report_codes.QDEVICE_DESTROY_SUCCESS,
                    {
                        "model": "net",
                    },
                ),
            ],
        )

    def test_used_not_forced(self, mock_net_destroy, mock_status):
        mock_status.return_value = 'Cluster "a_cluster":\n'

        assert_raise_library_error(
            lambda: lib.qdevice_destroy(self.lib_env, "net")
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.ERROR,
                    report_codes.QDEVICE_USED_BY_CLUSTERS,
                    {
                        "clusters": ["a_cluster"],
                    },
                    report_codes.FORCE,
                ),
            ],
        )

        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [],
        )
        mock_net_destroy.assert_not_called()

    def test_stop_failed(self, mock_net_destroy, mock_status):
        mock_status.return_value = ""
        self.lib_env.service_manager.stop.side_effect = StopServiceError(
            "test service", "test error"
        )

        assert_raise_library_error(
            lambda: lib.qdevice_destroy(self.lib_env, "net"),
            (
                severity.ERROR,
                reports.codes.SERVICE_ACTION_FAILED,
                {
                    "action": reports.const.SERVICE_ACTION_STOP,
                    "service": "test service",
                    "reason": "test error",
                    "node": "",
                    "instance": "",
                },
            ),
        )

        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.stop("corosync-qnetd")],
        )
        mock_net_destroy.assert_not_called()
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "instance": "",
                    },
                )
            ],
        )

    def test_disable_failed(self, mock_net_destroy, mock_status):
        mock_status.return_value = ""
        self.lib_env.service_manager.disable.side_effect = DisableServiceError(
            "test service", "test error"
        )

        assert_raise_library_error(
            lambda: lib.qdevice_destroy(self.lib_env, "net"),
            (
                severity.ERROR,
                reports.codes.SERVICE_ACTION_FAILED,
                {
                    "action": reports.const.SERVICE_ACTION_DISABLE,
                    "service": "test service",
                    "reason": "test error",
                    "node": "",
                    "instance": "",
                },
            ),
        )

        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [
                mock.call.stop("corosync-qnetd"),
                mock.call.disable("corosync-qnetd"),
            ],
        )
        mock_net_destroy.assert_not_called()
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
            ],
        )

    def test_destroy_failed(self, mock_net_destroy, mock_status):
        mock_status.return_value = ""
        mock_net_destroy.side_effect = LibraryError("mock_report_item")

        self.assertRaises(
            LibraryError, lambda: lib.qdevice_destroy(self.lib_env, "net")
        )

        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [
                mock.call.stop("corosync-qnetd"),
                mock.call.disable("corosync-qnetd"),
            ],
        )
        mock_net_destroy.assert_called_once_with()
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_DISABLE,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
            ],
        )


@mock.patch("pcs.lib.commands.qdevice.qdevice_net.qdevice_status_cluster_text")
@mock.patch("pcs.lib.commands.qdevice.qdevice_net.qdevice_status_generic_text")
@mock.patch.object(LibraryEnvironment, "cmd_runner", lambda self: "mock_runner")
class TestQdeviceNetStatusTextTest(QdeviceTestCase):
    def test_success(self, mock_status_generic, mock_status_cluster):
        mock_status_generic.return_value = "generic status info\n"
        mock_status_cluster.return_value = "cluster status info\n"

        self.assertEqual(
            lib.qdevice_status_text(self.lib_env, "net"),
            "generic status info\ncluster status info\n",
        )

        mock_status_generic.assert_called_once_with("mock_runner", False)
        mock_status_cluster.assert_called_once_with("mock_runner", None, False)

    def test_success_verbose(self, mock_status_generic, mock_status_cluster):
        mock_status_generic.return_value = "generic status info\n"
        mock_status_cluster.return_value = "cluster status info\n"

        self.assertEqual(
            lib.qdevice_status_text(self.lib_env, "net", verbose=True),
            "generic status info\ncluster status info\n",
        )

        mock_status_generic.assert_called_once_with("mock_runner", True)
        mock_status_cluster.assert_called_once_with("mock_runner", None, True)

    def test_success_cluster(self, mock_status_generic, mock_status_cluster):
        mock_status_generic.return_value = "generic status info\n"
        mock_status_cluster.return_value = "cluster status info\n"

        self.assertEqual(
            lib.qdevice_status_text(self.lib_env, "net", cluster="name"),
            "generic status info\ncluster status info\n",
        )

        mock_status_generic.assert_called_once_with("mock_runner", False)
        mock_status_cluster.assert_called_once_with(
            "mock_runner", "name", False
        )

    def test_error_generic_status(
        self, mock_status_generic, mock_status_cluster
    ):
        mock_status_generic.side_effect = LibraryError("mock_report_item")
        mock_status_cluster.return_value = "cluster status info\n"

        self.assertRaises(
            LibraryError, lambda: lib.qdevice_status_text(self.lib_env, "net")
        )

        mock_status_generic.assert_called_once_with("mock_runner", False)
        mock_status_cluster.assert_not_called()

    def test_error_cluster_status(
        self, mock_status_generic, mock_status_cluster
    ):
        mock_status_generic.return_value = "generic status info\n"
        mock_status_cluster.side_effect = LibraryError("mock_report_item")

        self.assertRaises(
            LibraryError, lambda: lib.qdevice_status_text(self.lib_env, "net")
        )

        mock_status_generic.assert_called_once_with("mock_runner", False)
        mock_status_cluster.assert_called_once_with("mock_runner", None, False)


class QdeviceNetEnableTest(QdeviceTestCase):
    # pylint: disable=no-member
    def test_success(self):
        lib.qdevice_enable(self.lib_env, "net")
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.enable("corosync-qnetd")],
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_ENABLE,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                )
            ],
        )

    def test_failed(self):
        self.lib_env.service_manager.enable.side_effect = EnableServiceError(
            "test service", "test error"
        )

        assert_raise_library_error(
            lambda: lib.qdevice_enable(self.lib_env, "net"),
            (
                severity.ERROR,
                reports.codes.SERVICE_ACTION_FAILED,
                {
                    "action": reports.const.SERVICE_ACTION_ENABLE,
                    "service": "test service",
                    "reason": "test error",
                    "node": "",
                    "instance": "",
                },
            ),
        )
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.enable("corosync-qnetd")],
        )


class QdeviceNetDisableTest(QdeviceTestCase):
    # pylint: disable=no-member
    def test_success(self):
        lib.qdevice_disable(self.lib_env, "net")
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.disable("corosync-qnetd")],
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_DISABLE,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                )
            ],
        )

    def test_failed(self):
        self.lib_env.service_manager.disable.side_effect = DisableServiceError(
            "test service", "test error"
        )

        assert_raise_library_error(
            lambda: lib.qdevice_disable(self.lib_env, "net"),
            (
                severity.ERROR,
                reports.codes.SERVICE_ACTION_FAILED,
                {
                    "action": reports.const.SERVICE_ACTION_DISABLE,
                    "service": "test service",
                    "reason": "test error",
                    "node": "",
                    "instance": "",
                },
            ),
        )
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.disable("corosync-qnetd")],
        )


@mock.patch("pcs.lib.corosync.qdevice_net.qdevice_initialized")
class QdeviceNetStartTest(QdeviceTestCase):
    # pylint: disable=no-member
    def test_success(self, mock_qdevice_initialized):
        mock_qdevice_initialized.return_value = True
        lib.qdevice_start(self.lib_env, "net")
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.start("corosync-qnetd")],
        )
        mock_qdevice_initialized.assert_called_once_with()
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_START,
                        "service": "quorum device",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_START,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
            ],
        )

    def test_failed(self, mock_qdevice_initialized):
        self.lib_env.service_manager.start.side_effect = StartServiceError(
            "test service", "test error"
        )
        mock_qdevice_initialized.return_value = True

        assert_raise_library_error(
            lambda: lib.qdevice_start(self.lib_env, "net"),
            (
                severity.ERROR,
                reports.codes.SERVICE_ACTION_FAILED,
                {
                    "action": reports.const.SERVICE_ACTION_START,
                    "service": "test service",
                    "reason": "test error",
                    "node": "",
                    "instance": "",
                },
            ),
        )
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.start("corosync-qnetd")],
        )
        mock_qdevice_initialized.assert_called_once_with()
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_START,
                        "service": "quorum device",
                        "instance": "",
                    },
                )
            ],
        )

    def test_qdevice_not_initialized(self, mock_qdevice_initialized):
        mock_qdevice_initialized.return_value = False

        assert_raise_library_error(
            lambda: lib.qdevice_start(self.lib_env, "net"),
            (
                severity.ERROR,
                report_codes.QDEVICE_NOT_INITIALIZED,
                {
                    "model": "net",
                },
            ),
        )
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [],
        )
        mock_qdevice_initialized.assert_called_once_with()


@mock.patch("pcs.lib.corosync.qdevice_net.qdevice_status_cluster_text")
class QdeviceNetStopTest(QdeviceTestCase):
    # pylint: disable=no-member
    def test_success_not_used(self, mock_status):
        mock_status.return_value = ""

        lib.qdevice_stop(self.lib_env, "net", proceed_if_used=False)

        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.stop("corosync-qnetd")],
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
            ],
        )

    def test_success_used_forced(self, mock_status):
        mock_status.return_value = 'Cluster "a_cluster":\n'

        lib.qdevice_stop(self.lib_env, "net", proceed_if_used=True)

        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.stop("corosync-qnetd")],
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.WARNING,
                    report_codes.QDEVICE_USED_BY_CLUSTERS,
                    {
                        "clusters": ["a_cluster"],
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "instance": "",
                    },
                ),
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                ),
            ],
        )

    def test_used_not_forced(self, mock_status):
        mock_status.return_value = 'Cluster "a_cluster":\n'

        assert_raise_library_error(
            lambda: lib.qdevice_stop(self.lib_env, "net", proceed_if_used=False)
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.ERROR,
                    report_codes.QDEVICE_USED_BY_CLUSTERS,
                    {
                        "clusters": ["a_cluster"],
                    },
                    report_codes.FORCE,
                ),
            ],
        )
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [],
        )

    def test_failed(self, mock_status):
        mock_status.return_value = ""
        self.lib_env.service_manager.stop.side_effect = StopServiceError(
            "test service", "test error"
        )

        assert_raise_library_error(
            lambda: lib.qdevice_stop(self.lib_env, "net"),
            (
                severity.ERROR,
                reports.codes.SERVICE_ACTION_FAILED,
                {
                    "action": reports.const.SERVICE_ACTION_STOP,
                    "service": "test service",
                    "reason": "test error",
                    "node": "",
                    "instance": "",
                },
            ),
        )
        self.assertEqual(
            self.lib_env.service_manager.method_calls,
            [mock.call.stop("corosync-qnetd")],
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_STARTED,
                    {
                        "action": reports.const.SERVICE_ACTION_STOP,
                        "service": "quorum device",
                        "instance": "",
                    },
                )
            ],
        )


@mock.patch("pcs.lib.external.kill_services")
@mock.patch.object(LibraryEnvironment, "cmd_runner", lambda self: "mock_runner")
class QdeviceNetKillTest(QdeviceTestCase):
    def test_success(self, mock_net_kill):
        lib.qdevice_kill(self.lib_env, "net")
        mock_net_kill.assert_called_once_with("mock_runner", ["corosync-qnetd"])
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    severity.INFO,
                    reports.codes.SERVICE_ACTION_SUCCEEDED,
                    {
                        "action": reports.const.SERVICE_ACTION_KILL,
                        "service": "quorum device",
                        "node": "",
                        "instance": "",
                    },
                )
            ],
        )

    def test_failed(self, mock_net_kill):
        mock_net_kill.side_effect = KillServicesError(
            ["test service"], "test error"
        )

        assert_raise_library_error(
            lambda: lib.qdevice_kill(self.lib_env, "net"),
            (
                severity.ERROR,
                reports.codes.SERVICE_ACTION_FAILED,
                {
                    "action": reports.const.SERVICE_ACTION_KILL,
                    "service": "test service",
                    "reason": "test error",
                    "node": "",
                    "instance": "",
                },
            ),
        )
        mock_net_kill.assert_called_once_with("mock_runner", ["corosync-qnetd"])


@mock.patch(
    "pcs.lib.commands.qdevice.qdevice_net.qdevice_sign_certificate_request"
)
@mock.patch.object(LibraryEnvironment, "cmd_runner", lambda self: "mock_runner")
class QdeviceNetSignCertificateRequestTest(QdeviceTestCase):
    def test_success(self, mock_qdevice_func):
        qdevice_func_input = "certificate request".encode("utf-8")
        qdevice_func_output = "signed certificate".encode("utf-8")
        mock_qdevice_func.return_value = qdevice_func_output
        cluster_name = "clusterName"

        self.assertEqual(
            base64.b64encode(qdevice_func_output),
            lib.qdevice_net_sign_certificate_request(
                self.lib_env, base64.b64encode(qdevice_func_input), cluster_name
            ),
        )

        mock_qdevice_func.assert_called_once_with(
            "mock_runner", qdevice_func_input, cluster_name
        )

    def test_bad_input(self, mock_qdevice_func):
        qdevice_func_input = "certificate request".encode("utf-8")
        cluster_name = "clusterName"

        assert_raise_library_error(
            lambda: lib.qdevice_net_sign_certificate_request(
                self.lib_env, qdevice_func_input, cluster_name
            ),
            (
                severity.ERROR,
                report_codes.INVALID_OPTION_VALUE,
                {
                    "option_name": "qnetd certificate request",
                    "option_value": qdevice_func_input,
                    "allowed_values": ["base64 encoded certificate"],
                    "cannot_be_empty": False,
                    "forbidden_characters": None,
                },
            ),
        )

        mock_qdevice_func.assert_not_called()


@mock.patch("pcs.lib.commands.qdevice.qdevice_net.client_setup")
@mock.patch.object(LibraryEnvironment, "cmd_runner", lambda self: "mock_runner")
class ClientNetSetupTest(QdeviceTestCase):
    def test_success(self, mock_qdevice_func):
        qdevice_func_input = "CA certificate".encode("utf-8")

        lib.client_net_setup(self.lib_env, base64.b64encode(qdevice_func_input))

        mock_qdevice_func.assert_called_once_with(
            "mock_runner", qdevice_func_input
        )

    def test_bad_input(self, mock_qdevice_func):
        qdevice_func_input = "CA certificate".encode("utf-8")

        assert_raise_library_error(
            lambda: lib.client_net_setup(self.lib_env, qdevice_func_input),
            (
                severity.ERROR,
                report_codes.INVALID_OPTION_VALUE,
                {
                    "option_name": "qnetd CA certificate",
                    "option_value": qdevice_func_input,
                    "allowed_values": ["base64 encoded certificate"],
                    "cannot_be_empty": False,
                    "forbidden_characters": None,
                },
            ),
        )

        mock_qdevice_func.assert_not_called()


@mock.patch(
    "pcs.lib.commands.qdevice.qdevice_net.client_import_certificate_and_key"
)
@mock.patch.object(LibraryEnvironment, "cmd_runner", lambda self: "mock_runner")
class ClientNetImportCertificateTest(QdeviceTestCase):
    def test_success(self, mock_qdevice_func):
        qdevice_func_input = "client certificate".encode("utf-8")

        lib.client_net_import_certificate(
            self.lib_env, base64.b64encode(qdevice_func_input)
        )

        mock_qdevice_func.assert_called_once_with(
            "mock_runner", qdevice_func_input
        )

    def test_bad_input(self, mock_qdevice_func):
        qdevice_func_input = "client certificate".encode("utf-8")

        assert_raise_library_error(
            lambda: lib.client_net_import_certificate(
                self.lib_env, qdevice_func_input
            ),
            (
                severity.ERROR,
                report_codes.INVALID_OPTION_VALUE,
                {
                    "option_name": "qnetd client certificate",
                    "option_value": qdevice_func_input,
                    "allowed_values": ["base64 encoded certificate"],
                    "cannot_be_empty": False,
                    "forbidden_characters": None,
                },
            ),
        )

        mock_qdevice_func.assert_not_called()


@mock.patch("pcs.lib.commands.qdevice.qdevice_net.client_destroy")
class ClientNetDestroyTest(QdeviceTestCase):
    def test_success(self, mock_qdevice_func):
        lib.client_net_destroy(self.lib_env)
        mock_qdevice_func.assert_called_once_with()



import re
from typing import (
    List,
    Optional,
    Union,
)

from pcs.common import reports
from pcs.common.reports.item import ReportItem, ReportItemList
from pcs.common.tools import timeout_to_seconds
from pcs.lib.errors import LibraryError


_BOOLEAN_TRUE = frozenset(["true", "on", "yes", "y", "1"])
_BOOLEAN_FALSE = frozenset(["false", "off", "no", "n", "0"])
_BOOLEAN = _BOOLEAN_TRUE | _BOOLEAN_FALSE
_ID_FIRST_CHAR_NOT_RE = re.compile("[^a-zA-Z_]")
_ID_REST_CHARS_NOT_RE = re.compile("[^a-zA-Z0-9_.-]")
SCORE_INFINITY = "INFINITY"


def is_boolean(val) -> bool:
    """
    Does pacemaker consider a value to be a boolean?
    See crm_is_true in pacemaker/lib/common/utils.c
    val checked value
    """
    return val.lower() in _BOOLEAN


def is_true(val) -> bool:
    """
    Does pacemaker consider a value to be true?
    See crm_is_true in pacemaker/lib/common/utils.c
    var checked value
    """
    return val.lower() in _BOOLEAN_TRUE


def is_false(val) -> bool:
    """
    Does pacemaker consider a value to be false?
    See crm_is_true in pacemaker/lib/common/utils.c
    var checked value
    """
    return val.lower() in _BOOLEAN_FALSE


def is_score(value: str) -> bool:
    if not value:
        return False
    unsigned_value = value[1:] if value[0] in ("+", "-") else value
    return unsigned_value == SCORE_INFINITY or unsigned_value.isdigit()


def get_valid_timeout_seconds(
    timeout_candidate: Union[str, int, None],
) -> Optional[int]:
    """
    Transform pacemaker style timeout to number of seconds, raise LibraryError
        on invalid timeout
    timeout_candidate timeout string or None
    """
    if timeout_candidate is None:
        return None
    wait_timeout = timeout_to_seconds(timeout_candidate)
    if wait_timeout is None:
        raise LibraryError(
            ReportItem.error(
                reports.messages.InvalidTimeoutValue(str(timeout_candidate))
            )
        )
    return wait_timeout


def validate_id(
    id_candidate: str,
    description: Optional[str] = None,
    reporter: Union[None, List, ReportItemList] = None,
):
    """
    Validate a pacemaker id, raise LibraryError on invalid id.

    id_candidate id's value
    description id's role description (default "id")
    """
    # see NCName definition
    # http://www.w3.org/TR/REC-xml-names/#NT-NCName
    # http://www.w3.org/TR/REC-xml/#NT-Name
    description = "id" if not description else description  # for mypy
    if not id_candidate:
        report_item = ReportItem.error(
            reports.messages.InvalidIdIsEmpty(description)
        )
        if reporter is None:
            # we check for None so it works with an empty list as well
            raise LibraryError(report_item)
        reporter.append(report_item)
        return
    if _ID_FIRST_CHAR_NOT_RE.match(id_candidate[0]):
        report_item = ReportItem.error(
            reports.messages.InvalidIdBadChar(
                id_candidate,
                description,
                id_candidate[0],
                True,
            )
        )
        if reporter is not None:
            reporter.append(report_item)
        else:
            raise LibraryError(report_item)
    for char in id_candidate[1:]:
        if _ID_REST_CHARS_NOT_RE.match(char):
            report_item = ReportItem.error(
                reports.messages.InvalidIdBadChar(
                    id_candidate,
                    description,
                    char,
                    False,
                )
            )
            if reporter is not None:
                reporter.append(report_item)
            else:
                raise LibraryError(report_item)


def sanitize_id(id_candidate, replacement=""):
    if not id_candidate:
        return id_candidate
    return "".join(
        [
            ""
            if _ID_FIRST_CHAR_NOT_RE.match(id_candidate[0])
            else id_candidate[0],
            _ID_REST_CHARS_NOT_RE.sub(replacement, id_candidate[1:]),
        ]
    )



from dataclasses import dataclass, field, replace as dc_replace
import re
from typing import (
    cast,
    Container,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Type,
    Union,
)

from lxml import etree
from lxml.etree import _Element

from pcs import settings
from pcs.common import const, pacemaker, reports
from pcs.common.interface.dto import DataTransferObject, meta
from pcs.common.reports.types import SeverityLevel
from pcs.common.str_tools import format_list
from pcs.common.tools import xml_fromstring
from pcs.lib import validate
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.pacemaker.values import is_true

# TODO: fix pylint issue
# pylint: disable=no-self-use

# TODO move to resource create / operation create
# Operation monitor is required always! No matter if --no-default-ops was
# entered or if agent does not specify it. See
# https://clusterlabs.org/pacemaker/doc/2.1/Pacemaker_Explained/html/resources.html#monitoring-resources-for-failure
_NECESSARY_CIB_ACTION_NAMES = {"monitor"}

# These are all standards valid in cib. To get a list of standards supported by
# pacemaker in local environment, use result of "crm_resource --list-standards".
_ALLOWED_STANDARDS = {
    "ocf",
    "lsb",
    "heartbeat",
    "stonith",
    "upstart",
    "service",
    "systemd",
    "nagios",
}

# TODO move to resource create / operation create
_DEFAULT_INTERVALS = {"monitor": "60s"}

STONITH_ACTION_REPLACED_BY = ["pcmk_off_action", "pcmk_reboot_action"]


### exceptions


class ResourceAgentError(Exception):
    # pylint: disable=super-init-not-called
    def __init__(self, agent_name: str, message: str = ""):
        self.agent = agent_name
        self.message = message


class UnableToGetAgentMetadata(ResourceAgentError):
    pass


class InvalidResourceAgentName(ResourceAgentError):
    pass


class InvalidStonithAgentName(ResourceAgentError):
    pass


### data classes


@dataclass(frozen=True)
class ResourceAgentName:
    standard: str
    provider: Optional[str]
    type: str

    @property
    def full_name(self) -> str:
        return ":".join(filter(None, [self.standard, self.provider, self.type]))


@dataclass(frozen=True)
class AgentParameterDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes

    # name of the parameter
    name: str
    # short description
    shortdesc: str
    # long description
    longdesc: str
    # data type of the parameter
    type: str
    # default value of the parameter
    default: Optional[str]
    # True if it is a required parameter, False otherwise
    required: bool
    # True if the parameter is meant for advanced users
    advanced: bool
    # True if the parameter is deprecated, False otherwise
    deprecated: bool
    # list of parameters deprecating this one
    deprecated_by: List[str]
    # name of a deprecated parameter obsoleted by this one
    obsoletes: Optional[str]
    # should the parameter's value be unique across same agent resources?
    unique: bool
    # pcs originated warning
    # TODO probably should not be part of the DTO, it should be produced by
    # some validator
    pcs_deprecated_warning: Optional[str]


@dataclass(frozen=True)
class AgentActionDto(DataTransferObject):
    # pylint: disable=too-many-instance-attributes

    # (start, stop, promote...), mandatory by both OCF 1.0 and 1.1
    name: str
    # mandatory by both OCF 1.0 and 1.1, sometimes not defined by agents
    timeout: Optional[str]
    # optional by both OCF 1.0 and 1.1
    interval: Optional[str]
    # optional by OCF 1.1
    # not allowed by OCF 1.0, defined in OCF 1.0 agents anyway
    role: Optional[str]
    # OCF name: 'start-delay', optional by both OCF 1.0 and 1.1
    start_delay: Optional[str] = field(metadata=meta(name="start-delay"))
    # optional by both OCF 1.0 and 1.1
    depth: Optional[str]
    # not allowed by any OCF, defined in OCF 1.0 agents anyway
    automatic: Optional[str]
    # not allowed by any OCF, defined in OCF 1.0 agents anyway
    on_target: Optional[str]
    # TODO remove, mirror of 'depth' only for pcs internal usage
    # pylint: disable=invalid-name
    OCF_CHECK_LEVEL: Optional[str]


@dataclass(frozen=True)
class AgentMetadataDto(DataTransferObject):
    name: str
    shortdesc: str
    longdesc: str
    parameters: List[AgentParameterDto]
    actions: List[AgentActionDto]
    # TODO move to resource create
    default_actions: List[AgentActionDto]


### complete intervals in operations


# TODO move to resource create / operation create
def _get_default_interval(operation_name: str) -> str:
    """
    Return default interval for given operation_name
    """
    return _DEFAULT_INTERVALS.get(operation_name, "0s")


# TODO move to resource create / operation create
def _complete_all_intervals(
    raw_operation_list: Iterable[AgentActionDto],
) -> List[AgentActionDto]:
    """
    Return a new list of operations with "interval" defined for all of them

    operation_list -- can include items with "interval" == None
    """
    operation_list = []
    for operation in raw_operation_list:
        if operation.interval is None:
            operation_list.append(
                dc_replace(
                    operation, interval=_get_default_interval(operation.name)
                )
            )
        else:
            operation_list.append(operation)
    return operation_list


### Agent wrapper / loader / facade classes


class Agent:
    """
    Base class for providing convinient access to an agent's metadata
    """

    _agent_type_label: str = "agent"

    def __init__(self, runner: CommandRunner):
        """
        create an instance which reads metadata by itself on demand
        """
        self._runner = runner
        self._metadata: Optional[_Element] = None

    def _get_name(self) -> str:
        raise NotImplementedError()

    def get_name_info(self) -> AgentMetadataDto:
        """
        Get structured agent's info, only name is populated
        """
        return AgentMetadataDto(self._get_name(), "", "", [], [], [])

    def get_full_info(self) -> AgentMetadataDto:
        """
        Get structured agent's info, all items are populated
        """
        return AgentMetadataDto(
            self._get_name(),
            self._get_shortdesc(),
            self._get_longdesc(),
            self._get_parameters(),
            # TODO move to resource create / operation create
            # and use _get_raw_actions instead (renamed to _get_actions)
            self.get_actions(),
            # TODO move to resource create / operation create
            self.get_cib_default_actions(),
        )

    def _get_shortdesc(self) -> str:
        """
        Get a short description of agent's purpose
        """
        return self._get_text_from_dom_element(
            self._get_metadata().find("shortdesc")
        ) or self._get_metadata().get("shortdesc", "")

    def _get_longdesc(self) -> str:
        """
        Get a long description of agent's purpose
        """
        return self._get_text_from_dom_element(
            self._get_metadata().find("longdesc")
        )

    def _get_parameters(self) -> List[AgentParameterDto]:
        """
        Get list of agent's parameters
        """
        params_element = self._get_metadata().find("parameters")
        if params_element is None:
            return []

        preprocessed_param_list = []
        deprecated_by_dict: Dict[str, Set[str]] = {}
        for param_el in params_element.iter("parameter"):
            param = self._get_parameter(param_el)
            if param is None:
                continue
            preprocessed_param_list.append(param)
            if param.obsoletes:
                obsoletes = param.obsoletes
                if not obsoletes in deprecated_by_dict:
                    deprecated_by_dict[obsoletes] = set()
                deprecated_by_dict[obsoletes].add(param.name)

        param_list = []
        for param in preprocessed_param_list:
            if param.name in deprecated_by_dict:
                param_list.append(
                    dc_replace(
                        param,
                        deprecated_by=sorted(deprecated_by_dict[param.name]),
                    )
                )
            else:
                param_list.append(param)

        return param_list

    def _get_parameter(
        self, parameter_element: _Element
    ) -> Optional[AgentParameterDto]:
        if parameter_element.get("name", None) is None:
            return None

        value_type = "string"
        default_value = None
        content_element = parameter_element.find("content")
        if content_element is not None:
            value_type = content_element.get("type", value_type)
            default_value = content_element.get("default", default_value)

        return AgentParameterDto(
            str(parameter_element.attrib["name"]),
            self._get_text_from_dom_element(
                parameter_element.find("shortdesc")
            ),
            self._get_text_from_dom_element(parameter_element.find("longdesc")),
            value_type,
            default_value,
            required=is_true(parameter_element.get("required", "0")),
            advanced=False,
            deprecated=is_true(parameter_element.get("deprecated", "0")),
            deprecated_by=[],
            obsoletes=parameter_element.get("obsoletes", None),
            unique=is_true(parameter_element.get("unique", "0")),
            pcs_deprecated_warning=None,
        )

    def _get_parameter_obsoleting_chains(self) -> Dict[str, List[str]]:
        """
        get a dict describing parameters obsoleting

        Each key is a parameter which obsoletes parameters but is not itself
        obsoleted by any other parameter. Values are lists of obsoleted
        parameters: the first one is obsoleted by the key, the second one is
        obsoleted by the first one and so on.
        """
        # In meta-data, each param can have 'obsoletes' attribute containing a
        # name of a single param obsoleted by the param in question. That means:
        # 1) a deprecated param can be obsoleted by more than one param
        # 2) a param can obsolete none or one param
        # 3) there may be a loop (A obsoletes B, B obsoletes A)
        # This info is crucial when dealing with obsoleting and deprecated
        # params and it is reflected in the following piece of code.

        # first, get simple obsoleting mapping
        new_old = {
            param.name: param.obsoletes
            for param in self._get_parameters()
            if param.obsoletes
        }

        chains = {}
        # for each param which is not obsoleted
        for new_param in new_old:
            # start a new chain
            deprecated_by_new = []
            current_new = new_param
            # while current parameter obsoletes anything
            while current_new in new_old:
                # get the obsoleted parameter
                old = new_old[current_new]
                # if there is a loop, break
                if old in deprecated_by_new or old == current_new:
                    break
                # add the obsoleted parameter to the chain
                deprecated_by_new.append(old)
                # check if the obsoleted parameter obsoletes anything
                current_new = old
            if deprecated_by_new:
                chains[new_param] = deprecated_by_new
        return chains

    def validate_parameters_create(
        self,
        parameters: Mapping[str, str],
        force: bool = False,
        # TODO remove this argument, see pcs.lib.cib.commands.remote_node.create
        # for details
        do_not_report_instance_attribute_server_exists: bool = False,
    ) -> reports.ReportItemList:
        # This is just a basic validation checking that required parameters are
        # set and all set parameters are known to an agent. Missing checks are:
        # 1. values checks - if a param is an integer, then "abc" is not valid
        # 2. warnings should be emitted when a deprecated param is set
        # 3. errors should be emitted when a deprecated parameter and a
        #    parameter obsoleting it are set at the same time
        # 4. possibly some other checks
        # All of these have been missing in pcs since ever (ad 1. agents have
        # never provided enough info for us to do such validations, ad 2. and
        # 3. there were no deprecated parameters before). The checks should be
        # implemented in agents themselves, so I'm not adding them now either.
        report_items = []

        # report unknown parameters
        report_items.extend(
            validate.NamesIn(
                {param.name for param in self._get_parameters()},
                option_type=self._agent_type_label,
                severity=reports.get_severity(reports.codes.FORCE, force),
            ).validate(parameters)
        )
        # TODO remove this "if", see pcs.lib.cib.commands.remote_node.create
        # for details
        if do_not_report_instance_attribute_server_exists:
            for report_item in report_items:
                if isinstance(report_item, reports.ReportItem) and isinstance(
                    report_item.message, reports.messages.InvalidOptions
                ):
                    report_msg = cast(
                        reports.messages.InvalidOptions, report_item.message
                    )
                    report_item.message = reports.messages.InvalidOptions(
                        report_msg.option_names,
                        sorted(
                            [
                                value
                                for value in report_msg.allowed
                                if value != "server"
                            ]
                        ),
                        report_msg.option_type,
                        report_msg.allowed_patterns,
                    )

        # report missing required parameters
        missing_parameters = self._find_missing_required_parameters(parameters)
        if missing_parameters:
            report_items.append(
                reports.ReportItem(
                    severity=self._validate_report_severity(force),
                    message=reports.messages.RequiredOptionsAreMissing(
                        sorted(missing_parameters),
                        self._agent_type_label,
                    ),
                )
            )

        return report_items

    def validate_parameters_update(
        self,
        current_parameters: Mapping[str, str],
        new_parameters: Mapping[str, str],
        force: bool = False,
    ) -> reports.ReportItemList:
        # This is just a basic validation checking that required parameters are
        # set and all set parameters are known to an agent. Missing checks are:
        # 1. values checks - if a param is an integer, then "abc" is not valid
        # 2. warnings should be emitted when a deprecated param is set
        # 3. errors should be emitted when a deprecated parameter and a
        #    parameter obsoleting it are set at the same time
        # 4. possibly some other checks
        # All of these have been missing in pcs since ever (ad 1. agents have
        # never provided enough info for us to do such validations, ad 2. and
        # 3. there were no deprecated parameters before). The checks should be
        # implemented in agents themselves, so I'm not adding them now either.
        report_items = []

        # get resulting set of agent's parameters
        final_parameters = dict(current_parameters)
        for name, value in new_parameters.items():
            if value:
                final_parameters[name] = value
            else:
                if name in final_parameters:
                    del final_parameters[name]

        # report unknown parameters
        report_items.extend(
            validate.NamesIn(
                {param.name for param in self._get_parameters()},
                option_type=self._agent_type_label,
                severity=reports.get_severity(reports.codes.FORCE, force),
            ).validate(
                # Do not report unknown parameters already set in the CIB. They
                # have been reported already when the were added to the CIB.
                {
                    name: value
                    for name, value in new_parameters.items()
                    if name not in current_parameters
                }
            )
        )

        # report missing or removed required parameters
        missing_parameters = self._find_missing_required_parameters(
            final_parameters
        )
        if missing_parameters:
            report_items.append(
                reports.ReportItem(
                    severity=self._validate_report_severity(force),
                    message=reports.messages.RequiredOptionsAreMissing(
                        sorted(missing_parameters),
                        self._agent_type_label,
                    ),
                )
            )

        return report_items

    @staticmethod
    def _validate_report_severity(force: bool) -> reports.ReportItemSeverity:
        return reports.ReportItemSeverity(
            level=(
                reports.ReportItemSeverity.WARNING
                if force
                else reports.ReportItemSeverity.ERROR
            ),
            force_code=reports.codes.FORCE if not force else None,
        )

    def _find_missing_required_parameters(
        self, parameters: Mapping[str, str]
    ) -> Set[str]:
        missing_parameters = set()
        obsoleting_chains = self._get_parameter_obsoleting_chains()
        for param in self._get_parameters():
            if not param.required or param.deprecated_by:
                # non-required params are never required
                # we require non-deprecated params preferentially
                continue
            if param.name in parameters:
                # the param is not missing
                continue
            # the param is missing, maybe a deprecated one is set instead?
            if param.name in obsoleting_chains:
                obsoleted_set_instead = False
                for obsoleted_name in obsoleting_chains[param.name]:
                    if obsoleted_name in parameters:
                        obsoleted_set_instead = True
                        break
                if obsoleted_set_instead:
                    continue
            missing_parameters.add(param.name)
        return missing_parameters

    def _get_raw_actions(self) -> List[AgentActionDto]:
        actions_element = self._get_metadata().find("actions")
        if actions_element is None:
            return []
        return [
            AgentActionDto(
                str(action.attrib["name"]),
                action.get("timeout"),
                action.get("interval"),
                pacemaker.role.get_value_primary(
                    const.PcmkRoleType(str(action.attrib["role"]))
                )
                if action.get("role", None) is not None
                else action.get("role"),
                action.get("start-delay"),
                action.get("depth"),
                action.get("automatic"),
                action.get("on_target"),
                None,  # ocf_check_level, TODO remove
            )
            for action in actions_element.iter("action")
            if action.get("name", None) is not None
        ]

    # TODO move to resource create / operation create
    def get_actions(self) -> List[AgentActionDto]:
        """
        Get list of agent's actions (operations)
        """
        action_list = []
        for action in self._get_raw_actions():
            if action.depth is not None:
                if action.depth == "0":
                    action_list.append(dc_replace(action, depth=None))
                else:
                    action_list.append(
                        dc_replace(action, OCF_CHECK_LEVEL=action.depth)
                    )
            else:
                action_list.append(action)
        return action_list

    # TODO move to resource create
    def _is_cib_default_action(self, action: AgentActionDto) -> bool:
        del action
        return False

    # TODO move to resource create
    def get_cib_default_actions(
        self, necessary_only: bool = False
    ) -> List[AgentActionDto]:
        """
        List actions that should be put to resource on its creation.
        Note that every action has at least attribute name.
        """
        action_list = [
            action
            for action in self.get_actions()
            if (necessary_only and action.name in _NECESSARY_CIB_ACTION_NAMES)
            or (not necessary_only and self._is_cib_default_action(action))
        ]

        for action_name in _NECESSARY_CIB_ACTION_NAMES:
            if action_name not in [action.name for action in action_list]:
                action_list.append(
                    AgentActionDto(
                        action_name,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                        None,
                    )
                )

        return _complete_all_intervals(action_list)

    def _get_metadata(self) -> _Element:
        """
        Return metadata DOM
        Raise UnableToGetAgentMetadata if agent doesn't exist or unable to get
            or parse its metadata
        """
        if self._metadata is None:
            self._metadata = self._parse_metadata(self._load_metadata())
        return self._metadata

    def _load_metadata(self) -> str:
        raise NotImplementedError()

    def _parse_metadata(self, metadata: str) -> _Element:
        try:
            dom = xml_fromstring(metadata)
            # TODO Majority of agents don't provide valid metadata, so we skip
            # the validation for now. We want to enable it once the schema
            # and/or agents are fixed.
            # When enabling this check for overrides in child classes.
            # if os.path.isfile(settings.agent_metadata_schema):
            #    etree.DTD(file=settings.agent_metadata_schema).assertValid(dom)
            return dom
        except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
            raise UnableToGetAgentMetadata(self._get_name(), str(e)) from e

    def _get_text_from_dom_element(self, element: Optional[_Element]) -> str:
        if element is None or element.text is None:
            return ""
        return str(element.text).strip()


class FakeAgentMetadata(Agent):
    def _get_name(self) -> str:
        raise NotImplementedError()

    def _load_metadata(self) -> str:
        raise NotImplementedError()


class FencedMetadata(FakeAgentMetadata):
    def _get_name(self) -> str:
        return "pacemaker-fenced"

    def _get_parameter(
        self, parameter_element: _Element
    ) -> Optional[AgentParameterDto]:
        parameter = super()._get_parameter(parameter_element)
        if parameter is None:
            return None
        # Metadata are written in such a way that a longdesc text is a
        # continuation of a shortdesc text.
        return dc_replace(
            parameter,
            longdesc=f"{parameter.shortdesc}\n{parameter.longdesc}".strip(),
            advanced=parameter.shortdesc.startswith("Advanced use only"),
        )

    def _load_metadata(self) -> str:
        stdout, stderr, dummy_retval = self._runner.run(
            [settings.pacemaker_fenced, "metadata"]
        )
        metadata = stdout.strip()
        if not metadata:
            raise UnableToGetAgentMetadata(self._get_name(), stderr.strip())
        return metadata


class CrmAgent(Agent):
    def __init__(self, runner: CommandRunner, name: str):
        super().__init__(runner)
        self._name_parts = self._prepare_name_parts(name)

    def _prepare_name_parts(self, name: str) -> ResourceAgentName:
        raise NotImplementedError()

    def _get_full_name(self) -> str:
        return self._name_parts.full_name

    # TODO remove, use DTO instead
    def get_standard(self) -> str:
        return self._name_parts.standard

    # TODO remove, use DTO instead
    def get_provider(self) -> Optional[str]:
        return self._name_parts.provider

    # TODO remove, use DTO instead
    def get_type(self) -> str:
        return self._name_parts.type

    # TODO this is used only in this module, put it into right place and make
    # it private
    def is_valid_metadata(self) -> bool:
        """
        If we are able to get metadata, we consider the agent existing and valid
        """
        # if the agent is valid, we do not need to load its metadata again
        try:
            self._get_metadata()
        except UnableToGetAgentMetadata:
            return False
        return True

    # TODO this is used only in this module, put it into right place and make
    # it private
    def validate_metadata(self) -> "CrmAgent":
        """
        Validate metadata by attempt to retrieve it.
        """
        self._get_metadata()
        return self

    def _load_metadata(self) -> str:
        env_path = ":".join(
            [
                # otherwise pacemaker cannot run RHEL fence agents to get their
                # metadata
                settings.fence_agent_binaries,
                # otherwise heartbeat and cluster-glue agents don't work
                "/bin",
                # otherwise heartbeat and cluster-glue agents don't work
                "/usr/bin",
            ]
        )
        stdout, stderr, retval = self._runner.run(
            [
                settings.crm_resource_binary,
                "--show-metadata",
                self._get_full_name(),
            ],
            env_extend={
                "PATH": env_path,
            },
        )
        if retval != 0:
            raise UnableToGetAgentMetadata(self._get_name(), stderr.strip())
        return stdout.strip()


class ResourceAgent(CrmAgent):
    """
    Provides convinient access to a resource agent's metadata
    """

    _agent_type_label = "resource"

    def _prepare_name_parts(self, name: str) -> ResourceAgentName:
        return _split_resource_agent_name(name)

    def _get_name(self) -> str:
        return self._get_full_name()

    def _get_parameters(self) -> List[AgentParameterDto]:
        parameters = super()._get_parameters()
        if self.get_standard() == "ocf" and (
            self.get_provider() in ("heartbeat", "pacemaker")
        ):
            trace_ra_found = False
            trace_file_found = False
            for param in parameters:
                param_name = param.name.lower()
                if param_name == "trace_ra":
                    trace_ra_found = True
                elif param_name == "trace_file":
                    trace_file_found = True
                if trace_file_found and trace_ra_found:
                    break

            if not trace_ra_found:
                shortdesc = (
                    "Set to 1 to turn on resource agent tracing"
                    " (expect large output)"
                )
                parameters.append(
                    AgentParameterDto(
                        "trace_ra",
                        shortdesc,
                        (
                            shortdesc + " The trace output will be saved to "
                            "trace_file, if set, or by default to"
                            " $HA_VARRUN/ra_trace/<type>/<id>.<action>."
                            "<timestamp> e.g. $HA_VARRUN/ra_trace/oracle/"
                            "db.start.2012-11-27.08:37:08"
                        ),
                        type="integer",
                        default="0",
                        required=False,
                        advanced=True,
                        deprecated=False,
                        deprecated_by=[],
                        obsoletes=None,
                        unique=False,
                        pcs_deprecated_warning=None,
                    )
                )
            if not trace_file_found:
                shortdesc = "Path to a file to store resource agent tracing log"
                parameters.append(
                    AgentParameterDto(
                        "trace_file",
                        shortdesc,
                        shortdesc,
                        type="string",
                        default="",
                        required=False,
                        advanced=True,
                        deprecated=False,
                        deprecated_by=[],
                        obsoletes=None,
                        unique=False,
                        pcs_deprecated_warning=None,
                    )
                )

        return parameters

    # TODO move to resource create
    def _is_cib_default_action(self, action: AgentActionDto) -> bool:
        # Copy all actions to the CIB even those not defined in the OCF standard
        # or pacemaker. This way even custom actions defined in a resource agent
        # will be copied to the CIB and run by pacemaker if they specify
        # an interval. See https://github.com/ClusterLabs/pcs/issues/132
        return action.name not in {
            # one-time action, not meant to be processed by pacemaker
            "meta-data",
            # deprecated alias of monitor
            "status",
            # one-time action, not meant to be processed by pacemaker
            "validate-all",
        }


class AbsentAgentMixin:
    def _load_metadata(self) -> str:
        return "<resource-agent/>"

    def validate_parameters_create(
        self,
        parameters: Mapping[str, str],
        force: bool = False,
        # TODO remove this argument, see pcs.lib.cib.commands.remote_node.create
        # for details
        do_not_report_instance_attribute_server_exists: bool = False,
    ) -> reports.ReportItemList:
        del parameters, force, do_not_report_instance_attribute_server_exists
        return []

    def validate_parameters_update(
        self,
        current_parameters: Mapping[str, str],
        new_parameters: Mapping[str, str],
        force: bool = False,
    ) -> reports.ReportItemList:
        del current_parameters, new_parameters, force
        return []


class AbsentResourceAgent(AbsentAgentMixin, ResourceAgent):
    pass


class StonithAgent(CrmAgent):
    """
    Provides convinient access to a stonith agent's metadata
    """

    _fenced_metadata: Optional[FencedMetadata] = None
    _agent_type_label: str = "stonith"

    # TODO This is only used in tests to make them work. Tests are calling this
    # to flush the cache and thus ensuring consistent behavior - the metadata
    # are always loaded in all tests. TODO: implement the cache in such a way
    # no such hack is needed al drop the cache altogether if architecture
    # allows that.
    @classmethod
    def clear_fenced_metadata_cache(cls) -> None:
        cls._fenced_metadata = None

    def _prepare_name_parts(self, name: str) -> ResourceAgentName:
        # pacemaker doesn't support stonith (nor resource) agents with : in type
        if ":" in name:
            raise InvalidStonithAgentName(name)
        return ResourceAgentName("stonith", None, name)

    def _get_name(self) -> str:
        return self.get_type()

    def _get_parameters(self) -> List[AgentParameterDto]:
        # TODO: fix pylint issue
        # pylint: disable=protected-access
        return (
            self._filter_parameters(super()._get_parameters())
            + self._get_fenced_metadata()._get_parameters()
        )

    def validate_parameters_create(
        self,
        parameters: Mapping[str, str],
        force: bool = False,
        # TODO remove this argument, see pcs.lib.cib.commands.remote_node.create
        # for details
        do_not_report_instance_attribute_server_exists: bool = False,
    ) -> reports.ReportItemList:
        report_list = super().validate_parameters_create(
            parameters, force=force
        )
        report_list.extend(
            self._validate_action_is_deprecated(parameters, force=force)
        )
        return report_list

    def validate_parameters_update(
        self,
        current_parameters: Mapping[str, str],
        new_parameters: Mapping[str, str],
        force: bool = False,
    ) -> reports.ReportItemList:
        report_list = super().validate_parameters_update(
            current_parameters, new_parameters, force=force
        )
        report_list.extend(
            self._validate_action_is_deprecated(new_parameters, force=force)
        )
        return report_list

    def _validate_action_is_deprecated(
        self, parameters: Mapping[str, str], force: bool = False
    ) -> reports.ReportItemList:
        if parameters.get("action", ""):
            return [
                reports.ReportItem(
                    self._validate_report_severity(force),
                    reports.messages.DeprecatedOption(
                        "action",
                        sorted(STONITH_ACTION_REPLACED_BY),
                        self._agent_type_label,
                    ),
                )
            ]
        return []

    def _filter_parameters(
        self, parameters: Iterable[AgentParameterDto]
    ) -> List[AgentParameterDto]:
        """
        Remove parameters that should not be available to the user.
        """
        # We don't allow the user to change these options which are only
        # intended to be used interactively on command line.
        remove_parameters = frozenset(("help", "version"))
        filtered_1 = []
        for param in parameters:
            if param.name in remove_parameters:
                continue
            if param.name == "action":
                # However we still need the user to be able to set 'action' due
                # to backward compatibility reasons. So we just mark it as not
                # required. We also move it to advanced params to indicate users
                # should not set it in most cases.
                filtered_1.append(
                    dc_replace(
                        param,
                        required=False,
                        advanced=True,
                        pcs_deprecated_warning=(
                            "Specifying 'action' is deprecated and not "
                            "necessary with current Pacemaker versions. "
                            "Use {0} instead."
                        ).format(format_list(STONITH_ACTION_REPLACED_BY)),
                    )
                )
            else:
                filtered_1.append(param)

        # 'port' parameter is required by a fence agent, but it is filled
        # automatically by pacemaker based on 'pcmk_host_map' or
        # 'pcmk_host_list' parameter (defined in fenced metadata). Pacemaker
        # marks 'port' parameter as not required for us.
        # It, however, doesn't mark any parameters deprecating 'port' as not
        # required se we must do so ourselves.
        port_related_params = set(["port"])
        new_deprecated = set(["port"])
        while new_deprecated:
            current_deprecated = new_deprecated
            new_deprecated = set()
            for param in filtered_1:
                if param.obsoletes in current_deprecated:
                    if param.name not in port_related_params:
                        new_deprecated.add(param.name)
                        port_related_params.add(param.name)
        filtered_2 = []
        for param in filtered_1:
            if param.name in port_related_params:
                filtered_2.append(dc_replace(param, required=False))
            else:
                filtered_2.append(param)

        return filtered_2

    def _get_fenced_metadata(self) -> FencedMetadata:
        # TODO: fix pylint issue
        # pylint: disable=protected-access
        if not self.__class__._fenced_metadata:
            self.__class__._fenced_metadata = FencedMetadata(self._runner)
        return self.__class__._fenced_metadata

    def get_provides_unfencing(self) -> bool:
        # self.get_actions returns an empty list
        # TODO: fix pylint issue
        # pylint: disable=protected-access
        for action in self._get_raw_actions():
            if (
                action.name == "on"
                and action.on_target == "1"
                and action.automatic == "1"
            ):
                return True
        return False

    # TODO move to resource create
    def _is_cib_default_action(self, action: AgentActionDto) -> bool:
        return action.name == "monitor"


class AbsentStonithAgent(AbsentAgentMixin, StonithAgent):
    def _get_parameters(self) -> List[AgentParameterDto]:
        return []


### agent name


def _split_resource_agent_name(full_agent_name: str) -> ResourceAgentName:
    # full_agent_name could be for example systemd:lvm2-pvscan@252:2
    # note that the second colon is not separator of provider and type
    match = re.match(
        "^(?P<standard>systemd|service):(?P<agent_type>[^:@]+@.*)$",
        full_agent_name,
    )
    if match:
        return ResourceAgentName(
            match.group("standard"), None, match.group("agent_type")
        )

    match = re.match(
        "^(?P<standard>[^:]+)(:(?P<provider>[^:]+))?:(?P<type>[^:]+)$",
        full_agent_name,
    )
    if not match:
        raise InvalidResourceAgentName(full_agent_name)

    standard = match.group("standard")
    provider = match.group("provider") if match.group("provider") else None
    agent_type = match.group("type")

    if standard not in _ALLOWED_STANDARDS:
        raise InvalidResourceAgentName(full_agent_name)

    if standard == "ocf" and not provider:
        raise InvalidResourceAgentName(full_agent_name)

    if standard != "ocf" and provider:
        raise InvalidResourceAgentName(full_agent_name)

    return ResourceAgentName(standard, provider, agent_type)


### get a list of standards, providers, agents


def list_resource_agents_standards(runner: CommandRunner) -> List[str]:
    """
    Return list of resource agents standards (ocf, lsb, ... ) on the local host
    """
    # retval is number of standards found
    stdout, dummy_stderr, dummy_retval = runner.run(
        [settings.crm_resource_binary, "--list-standards"]
    )
    # we are only interested in RESOURCE agents
    ignored_standards = frozenset(["stonith"])
    return _prepare_agent_list(stdout, ignored_standards)


def list_resource_agents_ocf_providers(runner: CommandRunner) -> List[str]:
    """
    Return list of resource agents ocf providers on the local host
    """
    # retval is number of providers found
    stdout, dummy_stderr, dummy_retval = runner.run(
        [settings.crm_resource_binary, "--list-ocf-providers"]
    )
    return _prepare_agent_list(stdout)


def list_resource_agents_standards_and_providers(
    runner: CommandRunner,
) -> List[str]:
    """
    Return list of all standard[:provider] on the local host
    """
    standards = list_resource_agents_standards(runner) + [
        f"ocf:{provider}"
        for provider in list_resource_agents_ocf_providers(runner)
    ]
    # do not list ocf resources twice
    try:
        standards.remove("ocf")
    except ValueError:
        pass
    return sorted(standards, key=str.lower)


def list_resource_agents(
    runner: CommandRunner, standard_provider: str
) -> List[str]:
    """
    Return list of resource agents for specified standard on the local host

    runner
    standard_provider -- standard[:provider], e.g. lsb, ocf, ocf:pacemaker
    """
    # retval is 0 on success, anything else when no agents found
    stdout, dummy_stderr, retval = runner.run(
        [settings.crm_resource_binary, "--list-agents", standard_provider]
    )
    if retval != 0:
        return []
    return _prepare_agent_list(stdout)


def list_stonith_agents(runner: CommandRunner) -> List[str]:
    """
    Return list of fence agents on the local host
    """
    # retval is 0 on success, anything else when no agents found
    stdout, dummy_stderr, retval = runner.run(
        [settings.crm_resource_binary, "--list-agents", "stonith"]
    )
    if retval != 0:
        return []
    ignored_agents = frozenset(
        [
            "fence_ack_manual",
            "fence_check",
            "fence_kdump_send",
            "fence_legacy",
            "fence_na",
            "fence_node",
            "fence_nss_wrapper",
            "fence_pcmk",
            "fence_sanlockd",
            "fence_tool",
            "fence_virtd",
            "fence_vmware_helper",
        ]
    )
    return _prepare_agent_list(stdout, ignored_agents)


def _prepare_agent_list(
    agents_string: str, ignored_items: Optional[Container] = None
) -> List[str]:
    """
    Parse list of agents / standards / providers / etc provided by crm_resource
    """
    ignored = ignored_items or frozenset([])
    result = [
        name
        for name in [line.strip() for line in agents_string.splitlines()]
        if name and name not in ignored
    ]
    return sorted(result, key=str.lower)


### get an Agent instance by agent's name


def _guess_resource_agent_full_name(
    runner: CommandRunner, search_agent_name: str
) -> List[ResourceAgent]:
    """
    List resource agents matching specified search term

    search_agent_name -- part of full agent name
    """
    search_lower = search_agent_name.lower()
    # list all possible names
    possible_names = []
    for std in list_resource_agents_standards_and_providers(runner):
        for agent in list_resource_agents(runner, std):
            if search_lower == agent.lower():
                possible_names.append(f"{std}:{agent}")
    # construct agent wrappers
    agent_candidates = [
        ResourceAgent(runner, agent) for agent in possible_names
    ]
    # check if the agent is valid
    return [agent for agent in agent_candidates if agent.is_valid_metadata()]


def _guess_exactly_one_resource_agent_full_name(
    runner: CommandRunner, search_agent_name: str
) -> ResourceAgent:
    """
    Get one resource agent matching specified search term or raise LibraryError

    search_agent_name -- last part of full agent name
    """
    agents = _guess_resource_agent_full_name(runner, search_agent_name)
    if not agents:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.AgentNameGuessFoundNone(search_agent_name)
            )
        )
    if len(agents) > 1:
        raise LibraryError(
            reports.ReportItem.error(
                reports.messages.AgentNameGuessFoundMoreThanOne(
                    search_agent_name,
                    sorted([agent.get_name_info().name for agent in agents]),
                )
            )
        )
    return agents[0]


def find_valid_resource_agent_by_name(
    report_processor: reports.ReportProcessor,
    runner: CommandRunner,
    name: str,
    allowed_absent: bool = False,
    absent_agent_supported: bool = True,
) -> Union[ResourceAgent, AbsentResourceAgent]:
    """
    Return instance of ResourceAgent corresponding to name

    report_processor -- tool for warning/info/error reporting
    runner -- tool for launching external commands
    name -- specifies a searched agent
    allowed_absent -- return a dummy class if agent not found
    absent_agent_supported -- should an error "agent not found" be forcible?
    """
    if ":" not in name:
        agent = _guess_exactly_one_resource_agent_full_name(runner, name)
        report_processor.report(
            reports.ReportItem.info(
                reports.messages.AgentNameGuessed(
                    name, agent.get_name_info().name
                )
            )
        )
        return agent

    return cast(
        ResourceAgent,
        _find_valid_agent_by_name(
            report_processor,
            runner,
            name,
            ResourceAgent,
            AbsentResourceAgent if allowed_absent else None,
            absent_agent_supported=absent_agent_supported,
        ),
    )


def find_valid_stonith_agent_by_name(
    report_processor: reports.ReportProcessor,
    runner: CommandRunner,
    name: str,
    allowed_absent: bool = False,
    absent_agent_supported: bool = True,
) -> Union[StonithAgent, AbsentStonithAgent]:
    """
    Return instance of StonithAgent corresponding to name

    report_processor -- tool for warning/info/error reporting
    runner -- tool for launching external commands
    name -- specifies a searched agent
    allowed_absent -- return a dummy class if agent not found
    absent_agent_supported -- should an error "agent not found" be forcible?
    """
    return cast(
        StonithAgent,
        _find_valid_agent_by_name(
            report_processor,
            runner,
            name,
            StonithAgent,
            AbsentStonithAgent if allowed_absent else None,
            absent_agent_supported=absent_agent_supported,
        ),
    )


def _find_valid_agent_by_name(
    report_processor: reports.ReportProcessor,
    runner: CommandRunner,
    name: str,
    present_agent_class: Type[CrmAgent],
    absent_agent_class: Optional[
        Type[Union[AbsentResourceAgent, AbsentStonithAgent]]
    ],
    absent_agent_supported: bool = True,
) -> CrmAgent:
    try:
        return present_agent_class(runner, name).validate_metadata()
    except (InvalidResourceAgentName, InvalidStonithAgentName) as e:
        raise LibraryError(resource_agent_error_to_report_item(e)) from e
    except UnableToGetAgentMetadata as e:
        if not absent_agent_supported:
            raise LibraryError(resource_agent_error_to_report_item(e)) from e

        if not absent_agent_class:
            raise LibraryError(
                resource_agent_error_to_report_item(e, forceable=True)
            ) from e

        report_processor.report(
            resource_agent_error_to_report_item(
                e,
                severity=reports.ReportItemSeverity.WARNING,
            )
        )

        return absent_agent_class(runner, name)


def resource_agent_error_to_report_item(
    e: ResourceAgentError,
    severity: SeverityLevel = reports.ReportItemSeverity.ERROR,
    forceable: bool = False,
) -> reports.ReportItem:
    """
    Transform ResourceAgentError to ReportItem
    """
    force = None
    if e.__class__ == UnableToGetAgentMetadata:
        if severity == reports.ReportItemSeverity.ERROR and forceable:
            force = reports.codes.FORCE
        return reports.ReportItem(
            severity=reports.ReportItemSeverity(severity, force),
            message=reports.messages.UnableToGetAgentMetadata(
                e.agent, e.message
            ),
        )
    if e.__class__ == InvalidResourceAgentName:
        return reports.ReportItem.error(
            reports.messages.InvalidResourceAgentName(e.agent)
        )
    if e.__class__ == InvalidStonithAgentName:
        return reports.ReportItem.error(
            reports.messages.InvalidStonithAgentName(e.agent)
        )
    raise e



import re
from typing import (
    cast,
    List,
    Set,
    Pattern,
)

from lxml.etree import (
    _Element,
    _ElementTree,
)

from pcs.common import (
    const,
    reports,
)
from pcs.common.reports import codes as report_codes
from pcs.common.reports.item import ReportItem, ReportItemList
from pcs.common.tools import Version
from pcs.lib.cib import sections
from pcs.lib.errors import LibraryError
from pcs.lib.pacemaker.values import (
    sanitize_id,
    validate_id,
)
from pcs.lib.xml_tools import get_root, get_sub_element

_VERSION_FORMAT = r"(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<rev>\d+))?$"


class IdProvider:
    """
    Book ids for future use in the CIB and generate new ids accordingly
    """

    def __init__(self, cib_element: _Element):
        """
        cib_element -- any element of the xml to check against
        """
        self._cib = get_root(cib_element)
        self._booked_ids: Set[str] = set()

    def allocate_id(self, proposed_id: str) -> str:
        """
        Generate a new unique id based on the proposal and keep track of it

        string proposed_id -- requested id
        """
        final_id = find_unique_id(self._cib, proposed_id, self._booked_ids)
        self._booked_ids.add(final_id)
        return final_id

    def book_ids(self, *id_list: str) -> ReportItemList:
        """
        Check if the ids are not already used and reserve them for future use
        """
        reported_ids = set()
        report_list = []
        for _id in id_list:
            if _id in reported_ids:
                continue
            if _id in self._booked_ids or does_id_exist(self._cib, _id):
                report_list.append(
                    ReportItem.error(reports.messages.IdAlreadyExists(_id))
                )
                reported_ids.add(_id)
                continue
            self._booked_ids.add(_id)
        return report_list


class ElementSearcher:
    """
    Search for an element, allow to book its id if not found, provide reports

    Usage:
    es = ElementSearcher(element_tag, element_id, context_element)
    if es.element_found():
        # the element has been found
        element = es.get_element()
    # if you want to create the element when it does not exist
    elif es.validate_book_id(id_provider)
        element = Element(element_id)
    else:
        raise LibraryError(es.get_errors())
    """

    def __init__(
        self, tags, element_id, context_element, element_type_desc=None
    ):
        """
        string|iterable tags -- a tag (string) or tags (iterable) to look for
        string element_id -- an id to look for
        etree.Element context_element -- an element to look in
        string|iterable element_type_desc -- element types for reports, tags
            if not specified
        """
        self._executed = False
        self._element = None
        self._element_id = element_id
        self._context_element = context_element
        self._tag_list = [tags] if isinstance(tags, str) else tags
        self._expected_types = self._prepare_expected_types(element_type_desc)
        self._book_errors = None

    def _prepare_expected_types(self, element_type_desc):
        if element_type_desc is None:
            return self._tag_list
        if isinstance(element_type_desc, str):
            return [element_type_desc]
        return element_type_desc

    def element_found(self):
        if not self._executed:
            self._execute()
        return self._element is not None

    def get_element(self):
        if not self._executed:
            self._execute()
        return self._element

    def validate_book_id(self, id_provider, id_description="id"):
        """
        Book element_id in the id_provider, return True if success
        """
        self._book_errors = []
        validate_id(
            self._element_id,
            description=id_description,
            reporter=self._book_errors,
        )
        if not self._book_errors:
            self._book_errors += id_provider.book_ids(self._element_id)
        return len(self._book_errors) < 1

    def get_errors(self):
        """
        Report why the element has not been found or booking its id failed
        """
        if self.element_found() or (
            self._book_errors is not None and not self._book_errors
        ):
            raise AssertionError(
                "Improper usage: cannot report errors when there are none"
            )

        element_list = get_configuration_elements_by_id(
            self._context_element, self._element_id
        )

        if element_list:
            element = element_list[0]
            if element.tag in self._tag_list:
                return [
                    ReportItem.error(
                        reports.messages.ObjectWithIdInUnexpectedContext(
                            element.tag,
                            self._element_id,
                            self._context_element.tag,
                            self._context_element.attrib.get("id", ""),
                        )
                    )
                ]
            return [
                ReportItem.error(
                    reports.messages.IdBelongsToUnexpectedType(
                        self._element_id,
                        expected_types=self._expected_types,
                        current_type=element.tag,
                    )
                )
            ]
        if self._book_errors is None:
            return [
                ReportItem.error(
                    reports.messages.IdNotFound(
                        self._element_id,
                        sorted(self._expected_types),
                        self._context_element.tag,
                        self._context_element.attrib.get("id", ""),
                    )
                )
            ]
        return self._book_errors

    def _execute(self):
        self._executed = True
        for tag in self._tag_list:
            element_list = self._context_element.xpath(
                ".//*[local-name()=$tag_name and @id=$element_id]",
                tag_name=tag,
                element_id=self._element_id,
            )
            if element_list:
                self._element = element_list[0]
                return


def get_configuration_elements_by_id(
    tree: _Element, check_id: str
) -> List[_Element]:
    """
    Return any configuration elements (not in status section of cib) with value
    of attribute id specified as 'check_id'; skip any and all elements having id
    attribute which does not actually serve as an id.

    tree -- any element in xml tree, whole tree (not only its subtree) will be
        searched
    check_id -- id to find
    """
    # do not search in /cib/status, it may contain references to previously
    # existing and deleted resources and thus preventing creating them again

    # pacemaker creates an implicit resource for the pacemaker_remote
    # connection, which will be named the same as the value of the remote-node
    # attribute of the explicit resource. So the value of nvpair named
    # "remote-node" is considered to be id
    return cast(
        List[_Element],
        get_root(tree).xpath(
            """
            (
                /cib/*[name()!="status"]
                |
                /*[name()!="cib"]
            )
            //*[
                (
                    name()!="acl_target"
                    and
                    name()!="role"
                    and
                    name()!="obj_ref"
                    and
                    name()!="resource_ref"
                    and
                    @id=$check_id
                ) or (
                    name()="primitive"
                    and
                    meta_attributes[
                        nvpair[
                            @name="remote-node"
                            and
                            @value=$check_id
                        ]
                    ]
                )
            ]
            """,
            check_id=check_id,
        ),
    )


# DEPRECATED, use IdProvider instead
def does_id_exist(tree, check_id):
    """
    Checks to see if id exists in the xml dom passed
    tree cib -- etree node
    check_id -- id to check
    """
    return len(get_configuration_elements_by_id(tree, check_id)) > 0


# DEPRECATED, use IdProvider instead
def validate_id_does_not_exist(tree, _id):
    """
    tree cib etree node
    """
    if does_id_exist(tree, _id):
        raise LibraryError(
            ReportItem.error(reports.messages.IdAlreadyExists(_id))
        )


# DEPRECATED, use IdProvider instead
def find_unique_id(tree, check_id, reserved_ids=None):
    """
    Returns check_id if it doesn't exist in the dom, otherwise it adds
    an integer to the end of the id and increments it until a unique id is found
    etree tree -- cib etree node
    string check_id -- id to check
    iterable reserved_ids -- ids to think about as already used
    """
    if not reserved_ids:
        reserved_ids = set()
    counter = 1
    temp_id = check_id
    while temp_id in reserved_ids or does_id_exist(tree, temp_id):
        temp_id = "{0}-{1}".format(check_id, counter)
        counter += 1
    return temp_id


# DEPRECATED, use ElementSearcher instead
def find_element_by_tag_and_id(
    tag, context_element, element_id, none_if_id_unused=False, id_types=None
):
    """
    Return element with given tag and element_id under context_element. When
    element does not exists raises LibraryError or return None if specified in
    none_if_id_unused.

    etree.Element(Tree) context_element is part of tree for element scan
    string|list tag is expected tag (or list of tags) of search element
    string element_id is id of search element
    bool none_if_id_unused if the element is not found then return None if True
        or raise a LibraryError if False
    list id_types optional list of descriptions for id / expected types of id
    """
    searcher = ElementSearcher(
        tag, element_id, context_element, element_type_desc=id_types
    )
    if searcher.element_found():
        return searcher.get_element()
    report_list = searcher.get_errors()
    if not none_if_id_unused:
        raise LibraryError(*report_list)
    filtered_reports = [
        report_item
        for report_item in report_list
        if report_item.message.code != report_codes.ID_NOT_FOUND
    ]
    if filtered_reports:
        raise LibraryError(*filtered_reports)
    return None


def create_subelement_id(
    context_element: _Element, suffix: str, id_provider: IdProvider
) -> str:
    proposed_id = sanitize_id(
        "{0}-{1}".format(context_element.get("id", context_element.tag), suffix)
    )
    return id_provider.allocate_id(proposed_id)


# DEPRECATED
# use ElementSearcher, IdProvider or pcs.lib.validate.ValueId instead
def check_new_id_applicable(tree, description, _id):
    validate_id(_id, description)
    validate_id_does_not_exist(tree, _id)


def get_configuration(tree):
    """
    Return 'configuration' element from tree, raise LibraryError if missing
    tree cib etree node
    """
    return sections.get(tree, sections.CONFIGURATION)


def get_acls(tree):
    """
    Return 'acls' element from tree, create a new one if missing
    tree cib etree node
    """
    return sections.get(tree, sections.ACLS)


def get_alerts(tree):
    """
    Return 'alerts' element from tree, create a new one if missing
    tree -- cib etree node
    """
    return sections.get(tree, sections.ALERTS)


def get_constraints(tree: _Element) -> _Element:
    """
    Return 'constraint' element from tree
    tree cib etree node
    """
    return sections.get(tree, sections.CONSTRAINTS)


def get_crm_config(tree: _Element) -> _Element:
    """
    Return 'crm_config' element from tree, raise LibraryError if missing

    tree -- cib etree node
    """
    return sections.get(tree, sections.CRM_CONFIG)


def get_fencing_topology(tree):
    """
    Return the 'fencing-topology' element from the tree
    tree -- cib etree node
    """
    return sections.get(tree, sections.FENCING_TOPOLOGY)


def get_nodes(tree):
    """
    Return 'nodes' element from the tree
    tree cib etree node
    """
    return sections.get(tree, sections.NODES)


def get_resources(tree: _Element) -> _Element:
    """
    Return the 'resources' element from the tree

    tree -- cib etree node
    """
    return sections.get(tree, sections.RESOURCES)


def get_status(tree):
    """
    Return the 'status' element from the tree
    tree -- cib etree node
    """
    return get_sub_element(tree, "status")


def get_tags(tree: _Element) -> _Element:
    """
    Return 'tags' element from tree, create a new one if missing
    tree -- cib etree node
    """
    return sections.get(tree, sections.TAGS)


def _get_cib_version(
    cib: _ElementTree, attribute: str, regexp: Pattern
) -> Version:
    version = cib.getroot().get(attribute)
    if version is None:
        raise LibraryError(
            ReportItem.error(
                reports.messages.CibLoadErrorBadFormat(
                    f"the attribute '{attribute}' of the element 'cib' "
                    "is missing"
                )
            )
        )
    match = regexp.match(version)
    if not match:
        raise LibraryError(
            ReportItem.error(
                reports.messages.CibLoadErrorBadFormat(
                    f"the attribute '{attribute}' of the element 'cib' has "
                    f"an invalid value: '{version}'"
                )
            )
        )
    return Version(
        int(match.group("major")),
        int(match.group("minor")),
        int(match.group("rev")) if match.group("rev") else None,
    )


def get_pacemaker_version_by_which_cib_was_validated(cib: _Element) -> Version:
    """
    Return version of pacemaker which validated specified cib as tree.
    Version is returned as an instance of pcs.common.tools.Version.
    Raises LibraryError on any failure.

    cib -- cib etree
    """
    return _get_cib_version(
        cib.getroottree(),
        "validate-with",
        re.compile(r"pacemaker-{0}".format(_VERSION_FORMAT)),
    )


def get_cib_crm_feature_set(cib: _Element) -> Version:
    """
    Return crm_feature_set as pcs.common.tools.Version or raise LibraryError

    etree cib -- cib etree
    """
    return _get_cib_version(
        cib.getroottree(), "crm_feature_set", re.compile(_VERSION_FORMAT)
    )


def are_new_role_names_supported(cib: _Element) -> bool:
    return (
        get_pacemaker_version_by_which_cib_was_validated(cib)
        >= const.PCMK_NEW_ROLES_CIB_VERSION
    )



from pcs.common import reports
from pcs.common.reports import ReportProcessor
from pcs.common.reports.item import ReportItem
from pcs.common.services.interfaces import ServiceManagerInterface
from pcs.lib import sbd
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.communication import (
    qdevice as qdevice_com,
    qdevice_net as qdevice_net_com,
)
from pcs.lib.communication.tools import run_and_raise
from pcs.lib.corosync import (
    config_validators as corosync_conf_validators,
    live as corosync_live,
    qdevice_net,
    qdevice_client,
)
from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncConfFacade
from pcs.lib.node import get_existing_nodes_names


def get_config(lib_env):
    """
    Extract and return quorum configuration from corosync.conf
    lib_env LibraryEnvironment
    """
    cfg = lib_env.get_corosync_conf()
    device = None
    if cfg.has_quorum_device():
        (
            model,
            model_options,
            generic_options,
            heuristics_options,
        ) = cfg.get_quorum_device_settings()
        device = {
            "model": model,
            "model_options": model_options,
            "generic_options": generic_options,
            "heuristics_options": heuristics_options,
        }
    return {
        "options": cfg.get_quorum_options(),
        "device": device,
    }


def _check_if_atb_can_be_disabled(
    service_manager: ServiceManagerInterface,
    report_processor: ReportProcessor,
    corosync_conf: CorosyncConfFacade,
    was_enabled: bool,
    force: bool = False,
) -> None:
    """
    Check whenever auto_tie_breaker can be changed without affecting SBD.
    Raises LibraryError if change of ATB will affect SBD functionality.

    service_manager --
    report_processor -- report processor
    corosync_conf -- corosync conf facade
    was_enabled -- True if ATB was enabled, False otherwise
    force -- force change
    """
    if (
        was_enabled
        and not corosync_conf.is_enabled_auto_tie_breaker()
        and sbd.is_auto_tie_breaker_needed(service_manager, corosync_conf)
    ):
        report_processor.report(
            ReportItem(
                severity=reports.item.get_severity(
                    reports.codes.FORCE,
                    force,
                ),
                message=(
                    reports.messages.CorosyncQuorumAtbCannotBeDisabledDueToSbd()
                ),
            )
        )
        if report_processor.has_errors:
            raise LibraryError()


def set_options(
    lib_env: LibraryEnvironment,
    options,
    skip_offline_nodes=False,
    force=False,
):
    """
    Set corosync quorum options, distribute and reload corosync.conf if live

    LibraryEnvironment lib_env
    dict options -- quorum options
    bool skip_offline_nodes -- continue even if not all nodes are accessible
    bool force -- force changes
    """
    cfg = lib_env.get_corosync_conf()
    if lib_env.report_processor.report_list(
        corosync_conf_validators.update_quorum_options(
            options, cfg.has_quorum_device(), cfg.get_quorum_options()
        )
    ).has_errors:
        raise LibraryError()
    cfg.set_quorum_options(options)
    if lib_env.is_corosync_conf_live:
        _check_if_atb_can_be_disabled(
            lib_env.service_manager,
            lib_env.report_processor,
            cfg,
            cfg.is_enabled_auto_tie_breaker(),
            force,
        )
    lib_env.push_corosync_conf(cfg, skip_offline_nodes)


def status_text(lib_env):
    """
    Get quorum runtime status in plain text
    """
    try:
        return corosync_live.get_quorum_status_text(lib_env.cmd_runner())
    except corosync_live.QuorumStatusReadException as e:
        raise LibraryError(
            ReportItem.error(
                reports.messages.CorosyncQuorumGetStatusError(e.reason)
            )
        ) from e


def status_device_text(lib_env, verbose=False):
    """
    Get quorum device client runtime status in plain text
    bool verbose get more detailed output
    """
    return qdevice_client.get_status_text(lib_env.cmd_runner(), verbose)


def add_device(
    lib_env: LibraryEnvironment,
    model,
    model_options,
    generic_options,
    heuristics_options,
    force_model=False,
    force_options=False,
    skip_offline_nodes=False,
):
    # pylint: disable=too-many-locals
    """
    Add a quorum device to a cluster, distribute and reload configs if live

    string model -- quorum device model
    dict model_options -- model specific options
    dict generic_options -- generic quorum device options
    dict heuristics_options -- heuristics options
    bool force_model -- continue even if the model is not valid
    bool force_options -- continue even if options are not valid
    bool skip_offline_nodes -- continue even if not all nodes are accessible
    """
    cfg = lib_env.get_corosync_conf()
    if cfg.has_quorum_device():
        raise LibraryError(
            ReportItem.error(reports.messages.QdeviceAlreadyDefined())
        )

    report_processor = lib_env.report_processor
    report_processor.report_list(
        corosync_conf_validators.add_quorum_device(
            model,
            model_options,
            generic_options,
            heuristics_options,
            [node.nodeid for node in cfg.get_nodes()],
            force_model=force_model,
            force_options=force_options,
        )
    )

    if lib_env.is_corosync_conf_live:
        cluster_nodes_names, report_list = get_existing_nodes_names(
            cfg,
            # Pcs is unable to communicate with nodes missing names. It cannot
            # send new corosync.conf to them. That might break the cluster.
            # Hence we error out.
            error_on_missing_name=True,
        )
        report_processor.report_list(report_list)

    if report_processor.has_errors:
        raise LibraryError()

    cfg.add_quorum_device(
        model,
        model_options,
        generic_options,
        heuristics_options,
    )
    if cfg.is_quorum_device_heuristics_enabled_with_no_exec():
        lib_env.report_processor.report(
            ReportItem.warning(
                reports.messages.CorosyncQuorumHeuristicsEnabledWithNoExec()
            )
        )

    # First setup certificates for qdevice, then send corosync.conf to nodes.
    # If anything fails, nodes will not have corosync.conf with qdevice in it,
    # so there is no effect on the cluster.
    if lib_env.is_corosync_conf_live:
        target_factory = lib_env.get_node_target_factory()
        target_list = target_factory.get_target_list(
            cluster_nodes_names,
            skip_non_existing=skip_offline_nodes,
        )
        # Do model specific configuration.
        # If the model is not known to pcs and was forced, do not configure
        # anything else than corosync.conf, as we do not know what to do
        # anyway.
        if model == "net":
            qdevice_net.set_up_client_certificates(
                lib_env.cmd_runner(),
                lib_env.report_processor,
                lib_env.communicator_factory,
                # We are sure the "host" key is there, it has been validated
                # above.
                target_factory.get_target_from_hostname(model_options["host"]),
                cfg.get_cluster_name(),
                target_list,
                skip_offline_nodes,
            )

        lib_env.report_processor.report(
            ReportItem.info(
                reports.messages.ServiceActionStarted(
                    reports.const.SERVICE_ACTION_ENABLE, "corosync-qdevice"
                )
            )
        )
        com_cmd = qdevice_com.Enable(
            lib_env.report_processor, skip_offline_nodes
        )
        com_cmd.set_targets(target_list)
        run_and_raise(lib_env.get_node_communicator(), com_cmd)

    # everything set up, it's safe to tell the nodes to use qdevice
    lib_env.push_corosync_conf(cfg, skip_offline_nodes)

    # Now, when corosync.conf has been reloaded, we can start qdevice service.
    if lib_env.is_corosync_conf_live:
        lib_env.report_processor.report(
            ReportItem.info(
                reports.messages.ServiceActionStarted(
                    reports.const.SERVICE_ACTION_START, "corosync-qdevice"
                )
            )
        )
        com_cmd_start = qdevice_com.Start(
            lib_env.report_processor, skip_offline_nodes
        )
        com_cmd_start.set_targets(target_list)
        run_and_raise(lib_env.get_node_communicator(), com_cmd_start)


def update_device(
    lib_env: LibraryEnvironment,
    model_options,
    generic_options,
    heuristics_options,
    force_options=False,
    skip_offline_nodes=False,
):
    """
    Change quorum device settings, distribute and reload configs if live

    dict model_options -- model specific options
    dict generic_options -- generic quorum device options
    dict heuristics_options -- heuristics options
    bool force_options -- continue even if options are not valid
    bool skip_offline_nodes -- continue even if not all nodes are accessible
    """
    cfg = lib_env.get_corosync_conf()
    if not cfg.has_quorum_device():
        raise LibraryError(
            ReportItem.error(reports.messages.QdeviceNotDefined())
        )
    if lib_env.report_processor.report_list(
        corosync_conf_validators.update_quorum_device(
            cfg.get_quorum_device_model(),
            model_options,
            generic_options,
            heuristics_options,
            [node.nodeid for node in cfg.get_nodes()],
            force_options=force_options,
        )
    ).has_errors:
        raise LibraryError()
    cfg.update_quorum_device(model_options, generic_options, heuristics_options)
    if cfg.is_quorum_device_heuristics_enabled_with_no_exec():
        lib_env.report_processor.report(
            ReportItem.warning(
                reports.messages.CorosyncQuorumHeuristicsEnabledWithNoExec()
            )
        )
    lib_env.push_corosync_conf(cfg, skip_offline_nodes)


def remove_device_heuristics(lib_env, skip_offline_nodes=False):
    """
    Stop using quorum device heuristics, distribute and reload configs if live

    bool skip_offline_nodes -- continue even if not all nodes are accessible
    """
    cfg = lib_env.get_corosync_conf()
    if not cfg.has_quorum_device():
        raise LibraryError(
            ReportItem.error(reports.messages.QdeviceNotDefined())
        )
    cfg.remove_quorum_device_heuristics()
    lib_env.push_corosync_conf(cfg, skip_offline_nodes)


def remove_device(lib_env: LibraryEnvironment, skip_offline_nodes=False):
    """
    Stop using quorum device, distribute and reload configs if live
    skip_offline_nodes continue even if not all nodes are accessible
    """
    cfg = lib_env.get_corosync_conf()
    if not cfg.has_quorum_device():
        raise LibraryError(
            ReportItem.error(reports.messages.QdeviceNotDefined())
        )
    model = cfg.get_quorum_device_model()
    cfg.remove_quorum_device()

    if lib_env.is_corosync_conf_live:
        report_processor = lib_env.report_processor
        # get nodes for communication
        cluster_nodes_names, report_list = get_existing_nodes_names(
            cfg,
            # Pcs is unable to communicate with nodes missing names. It cannot
            # send new corosync.conf to them. That might break the cluster.
            # Hence we error out.
            error_on_missing_name=True,
        )
        if report_processor.report_list(report_list).has_errors:
            raise LibraryError()
        target_list = lib_env.get_node_target_factory().get_target_list(
            cluster_nodes_names,
            skip_non_existing=skip_offline_nodes,
        )
        # fix quorum options for SBD to work properly
        if sbd.atb_has_to_be_enabled(lib_env.service_manager, cfg):
            lib_env.report_processor.report(
                ReportItem.warning(
                    reports.messages.CorosyncQuorumAtbWillBeEnabledDueToSbd()
                )
            )
            cfg.set_quorum_options({"auto_tie_breaker": "1"})

        # disable qdevice
        lib_env.report_processor.report(
            ReportItem.info(
                reports.messages.ServiceActionStarted(
                    reports.const.SERVICE_ACTION_DISABLE, "corosync-qdevice"
                )
            )
        )
        com_cmd_disable = qdevice_com.Disable(
            lib_env.report_processor, skip_offline_nodes
        )
        com_cmd_disable.set_targets(target_list)
        run_and_raise(lib_env.get_node_communicator(), com_cmd_disable)
        # stop qdevice
        lib_env.report_processor.report(
            ReportItem.info(
                reports.messages.ServiceActionStarted(
                    reports.const.SERVICE_ACTION_STOP, "corosync-qdevice"
                )
            )
        )
        com_cmd_stop = qdevice_com.Stop(
            lib_env.report_processor, skip_offline_nodes
        )
        com_cmd_stop.set_targets(target_list)
        run_and_raise(lib_env.get_node_communicator(), com_cmd_stop)
        # handle model specific configuration
        if model == "net":
            lib_env.report_processor.report(
                ReportItem.info(
                    reports.messages.QdeviceCertificateRemovalStarted()
                )
            )
            com_cmd_client_destroy = qdevice_net_com.ClientDestroy(
                lib_env.report_processor, skip_offline_nodes
            )
            com_cmd_client_destroy.set_targets(target_list)
            run_and_raise(
                lib_env.get_node_communicator(), com_cmd_client_destroy
            )

    lib_env.push_corosync_conf(cfg, skip_offline_nodes)


def set_expected_votes_live(lib_env, expected_votes):
    """
    set expected votes in live cluster to specified value
    numeric expected_votes desired value of expected votes
    """
    try:
        votes_int = int(expected_votes)
        if votes_int < 1:
            raise ValueError()
    except ValueError:
        raise LibraryError(
            ReportItem.error(
                reports.messages.InvalidOptionValue(
                    "expected votes", expected_votes, "positive integer"
                )
            )
        ) from None

    corosync_live.set_expected_votes(lib_env.cmd_runner(), votes_int)



import json
from functools import partial

from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.common.parse_args import prepare_options, group_by_keywords
from pcs.cli.reports.output import warn
from pcs.common.str_tools import indent

parse_cmd_sections = partial(group_by_keywords, implicit_first_group_key="main")


def ensure_only_allowed_options(parameter_dict, allowed_list):
    for arg, value in parameter_dict.items():
        if arg not in allowed_list:
            raise CmdLineInputError(
                "Unexpected parameter '{0}={1}'".format(arg, value)
            )


def alert_add(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    sections = parse_cmd_sections(argv, set(["options", "meta"]))
    main_args = prepare_options(sections["main"])
    ensure_only_allowed_options(main_args, ["id", "description", "path"])

    lib.alert.create_alert(
        main_args.get("id", None),
        main_args.get("path", None),
        prepare_options(sections["options"]),
        prepare_options(sections["meta"]),
        main_args.get("description", None),
    )


def alert_update(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    alert_id = argv[0]

    sections = parse_cmd_sections(argv[1:], set(["options", "meta"]))
    main_args = prepare_options(sections["main"])
    ensure_only_allowed_options(main_args, ["description", "path"])

    lib.alert.update_alert(
        alert_id,
        main_args.get("path", None),
        prepare_options(sections["options"]),
        prepare_options(sections["meta"]),
        main_args.get("description", None),
    )


def alert_remove(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    lib.alert.remove_alert(argv)


def recipient_add(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --force - allows not unique recipient values
    """
    modifiers.ensure_only_supported("-f", "--force")
    if len(argv) < 2:
        raise CmdLineInputError()

    alert_id = argv[0]

    sections = parse_cmd_sections(argv[1:], set(["options", "meta"]))
    main_args = prepare_options(sections["main"])
    ensure_only_allowed_options(main_args, ["description", "id", "value"])

    lib.alert.add_recipient(
        alert_id,
        main_args.get("value", None),
        prepare_options(sections["options"]),
        prepare_options(sections["meta"]),
        recipient_id=main_args.get("id", None),
        description=main_args.get("description", None),
        allow_same_value=modifiers.get("--force"),
    )


def recipient_update(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file (in lib wrapper)
      * --force - allows not unique recipient values
    """
    modifiers.ensure_only_supported("-f", "--force")
    if not argv:
        raise CmdLineInputError()

    recipient_id = argv[0]

    sections = parse_cmd_sections(argv[1:], set(["options", "meta"]))
    main_args = prepare_options(sections["main"])
    ensure_only_allowed_options(main_args, ["description", "value"])

    lib.alert.update_recipient(
        recipient_id,
        prepare_options(sections["options"]),
        prepare_options(sections["meta"]),
        recipient_value=main_args.get("value", None),
        description=main_args.get("description", None),
        allow_same_value=modifiers.get("--force"),
    )


def recipient_remove(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if not argv:
        raise CmdLineInputError()

    lib.alert.remove_recipient(argv)


def _nvset_to_str(nvset_obj):
    # TODO duplicite to pcs.resource._nvpairs_strings
    key_val = {
        nvpair_obj["name"]: nvpair_obj["value"] for nvpair_obj in nvset_obj
    }
    output = []
    for name, value in sorted(key_val.items()):
        if " " in value:
            value = f'"{value}"'
        output.append(f"{name}={value}")
    return " ".join(output)


def __description_attributes_to_str(obj):
    output = []
    if obj.get("description"):
        output.append("Description: {desc}".format(desc=obj["description"]))
    if obj.get("instance_attributes"):
        output.append(
            "Options: {attributes}".format(
                attributes=_nvset_to_str(obj["instance_attributes"])
            )
        )
    if obj.get("meta_attributes"):
        output.append(
            "Meta options: {attributes}".format(
                attributes=_nvset_to_str(obj["meta_attributes"])
            )
        )
    return output


def _alert_to_str(alert):
    content = []
    content.extend(__description_attributes_to_str(alert))

    recipients = []
    for recipient in alert.get("recipient_list", []):
        recipients.extend(_recipient_to_str(recipient))

    if recipients:
        content.append("Recipients:")
        content.extend(indent(recipients, 1))

    return [
        "Alert: {alert_id} (path={path})".format(
            alert_id=alert["id"], path=alert["path"]
        )
    ] + indent(content, 1)


def _recipient_to_str(recipient):
    return [
        "Recipient: {id} (value={value})".format(
            value=recipient["value"], id=recipient["id"]
        )
    ] + indent(__description_attributes_to_str(recipient), 1)


def print_alert_show(lib, argv, modifiers):
    warn(
        "This command is deprecated and will be removed. "
        "Please use 'pcs alert config' instead.",
        stderr=True,
    )
    return print_alert_config(lib, argv, modifiers)


def print_alert_config(lib, argv, modifiers):
    """
    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if argv:
        raise CmdLineInputError()
    print("\n".join(alert_config_lines(lib)))


def alert_config_lines(lib):
    lines = ["Alerts:"]
    alert_list = lib.alert.get_all_alerts()
    if alert_list:
        for alert in alert_list:
            lines.extend(indent(_alert_to_str(alert), 1))
    else:
        lines.append(" No alerts defined")
    return lines


def print_alerts_in_json(lib, argv, modifiers):
    """
    This is used only by pcsd, will be removed in new architecture

    Options:
      * -f - CIB file (in lib wrapper)
    """
    modifiers.ensure_only_supported("-f")
    if argv:
        raise CmdLineInputError()

    print(json.dumps(lib.alert.get_all_alerts()))



from typing import (
    cast,
    Iterable,
    List,
    Optional,
)

from pcs.cli.rule import (
    get_in_effect_label,
    rule_expression_dto_to_lines,
)
from pcs.common.pacemaker.nvset import CibNvsetDto
from pcs.common.str_tools import (
    format_name_value_list,
    format_optional,
    indent,
)
from pcs.common.types import (
    CibNvsetType,
    CibRuleInEffectStatus,
)


def nvset_dto_list_to_lines(
    nvset_dto_list: Iterable[CibNvsetDto],
    with_ids: bool = False,
    include_expired: bool = False,
    text_if_empty: Optional[str] = None,
) -> List[str]:
    if not nvset_dto_list:
        return [text_if_empty] if text_if_empty else []
    if not include_expired:
        nvset_dto_list = [
            nvset_dto
            for nvset_dto in nvset_dto_list
            if not nvset_dto.rule
            or nvset_dto.rule.in_effect != CibRuleInEffectStatus.EXPIRED
        ]
    return [
        line
        for nvset_dto in nvset_dto_list
        for line in nvset_dto_to_lines(nvset_dto, with_ids=with_ids)
    ]


def nvset_dto_to_lines(nvset: CibNvsetDto, with_ids: bool = False) -> List[str]:
    nvset_label = _nvset_type_to_label.get(nvset.type, "Options Set")
    in_effect_label = get_in_effect_label(nvset.rule) if nvset.rule else None
    heading_parts = [
        "{label}{in_effect}: {id}".format(
            label=nvset_label,
            in_effect=format_optional(in_effect_label, " ({})"),
            id=nvset.id,
        )
    ]
    if nvset.options:
        heading_parts.append(
            " ".join(format_name_value_list(sorted(nvset.options.items())))
        )

    lines = format_name_value_list(
        sorted([(nvpair.name, nvpair.value) for nvpair in nvset.nvpairs])
    )
    if nvset.rule:
        lines.extend(
            rule_expression_dto_to_lines(nvset.rule, with_ids=with_ids)
        )

    return [" ".join(heading_parts)] + indent(lines)


_nvset_type_to_label = {
    cast(str, CibNvsetType.INSTANCE): "Attributes",
    cast(str, CibNvsetType.META): "Meta Attrs",
}



"""
The intention is put there knowledge about cluster state structure.
Hide information about underlaying xml is desired too.
"""
from collections import defaultdict
from typing import Dict, Any

from lxml import etree

from pcs.common import (
    const,
    reports,
)
from pcs.common.reports.item import ReportItem
from pcs.lib.pacemaker.values import (
    is_false,
    is_true,
)
from pcs.lib.xml_tools import find_parent


class ResourceNotFound(Exception):
    pass


class _Attrs:
    def __init__(self, owner_name, attrib, required_attrs):
        """
        attrib lxml.etree._Attrib - wrapped attribute collection
        required_attrs dict of required atribute names object_name:xml_attribute
        """
        self.owner_name = owner_name
        self.attrib = attrib
        self.required_attrs = required_attrs

    def __getattr__(self, name):
        if name in self.required_attrs.keys():
            try:
                attr_specification = self.required_attrs[name]
                if isinstance(attr_specification, tuple):
                    attr_name, attr_transform = attr_specification
                    return attr_transform(self.attrib[attr_name])
                return self.attrib[attr_specification]
            except KeyError as e:
                raise AttributeError(
                    "Missing attribute '{0}' ('{1}' in source) in '{2}'".format(
                        name, self.required_attrs[name], self.owner_name
                    )
                ) from e

        raise AttributeError(
            "'{0}' does not declare attribute '{1}'".format(
                self.owner_name, name
            )
        )


class _Children:
    def __init__(self, owner_name, dom_part, children, sections):
        self.owner_name = owner_name
        self.dom_part = dom_part
        self.children = children
        self.sections = sections

    def __getattr__(self, name):
        if name in self.children.keys():
            element_name, wrapper = self.children[name]
            return [
                wrapper(element)
                for element in self.dom_part.findall(".//" + element_name)
            ]

        if name in self.sections.keys():
            element_name, wrapper = self.sections[name]
            return wrapper(self.dom_part.findall(".//" + element_name)[0])

        raise AttributeError(
            "'{0}' does not declare child or section '{1}'".format(
                self.owner_name, name
            )
        )


class _Element:
    # Note: not properly typed
    required_attrs: Dict[Any, Any] = {}
    # Note: not properly typed
    children: Dict[Any, Any] = {}
    # Note: not properly typed
    sections: Dict[Any, Any] = {}

    def __init__(self, dom_part: etree._Element):
        self.dom_part = dom_part
        self.attrs = _Attrs(
            self.__class__.__name__, self.dom_part.attrib, self.required_attrs
        )
        self.children_access = _Children(
            self.__class__.__name__,
            self.dom_part,
            self.children,
            self.sections,
        )

    def __getattr__(self, name):
        return getattr(self.children_access, name)


class _SummaryNodes(_Element):
    required_attrs = {
        "count": ("number", int),
    }


class _SummaryResources(_Element):
    required_attrs = {
        "count": ("number", int),
    }


class _SummarySection(_Element):
    sections = {
        "nodes": ("nodes_configured", _SummaryNodes),
        "resources": ("resources_configured", _SummaryResources),
    }


class _Node(_Element):
    required_attrs = {
        "id": "id",
        "name": "name",
        "type": "type",
        "online": ("online", is_true),
        "standby": ("standby", is_true),
        "standby_onfail": ("standby_onfail", is_true),
        "maintenance": ("maintenance", is_true),
        "pending": ("pending", is_true),
        "unclean": ("unclean", is_true),
        "shutdown": ("shutdown", is_true),
        "expected_up": ("expected_up", is_true),
        "is_dc": ("is_dc", is_true),
        "resources_running": ("resources_running", int),
    }


class _NodeSection(_Element):
    children = {
        "nodes": ("node", _Node),
    }


class ClusterState(_Element):
    sections = {
        "summary": ("summary", _SummarySection),
        "node_section": ("nodes", _NodeSection),
    }


def _id_xpath_predicate(resource_id):
    return """(@id="{0}" or starts-with(@id, "{0}:"))""".format(resource_id)


def _get_primitives_for_state_check(
    cluster_state, resource_id, expected_running
):
    primitives = cluster_state.xpath(
        """
        .//resource[{predicate_id}]
        |
        .//group[{predicate_id}]/resource[{predicate_position}]
        |
        .//clone[@id="{id}"]/resource
        |
        .//clone[@id="{id}"]/group/resource[{predicate_position}]
        |
        .//bundle[@id="{id}"]/replica/resource
    """.format(
            id=resource_id,
            predicate_id=_id_xpath_predicate(resource_id),
            predicate_position=("last()" if expected_running else "1"),
        )
    )
    return [
        element
        for element in primitives
        if not is_true(element.attrib.get("failed", ""))
    ]


def _get_primitive_roles_with_nodes(primitive_el_list):
    # Clone resources are represented by multiple primitive elements.
    roles_with_nodes = defaultdict(set)
    for resource_element in primitive_el_list:
        if resource_element.attrib["role"] in const.PCMK_ROLES_RUNNING:
            roles_with_nodes[resource_element.attrib["role"]].update(
                [
                    node.attrib["name"]
                    for node in resource_element.findall(".//node")
                ]
            )
    return {role: sorted(nodes) for role, nodes in roles_with_nodes.items()}


def get_resource_state(cluster_state, resource_id):
    return _get_primitive_roles_with_nodes(
        _get_primitives_for_state_check(
            cluster_state, resource_id, expected_running=True
        )
    )


def info_resource_state(cluster_state, resource_id):
    roles_with_nodes = get_resource_state(cluster_state, resource_id)
    if not roles_with_nodes:
        return ReportItem.info(reports.messages.ResourceDoesNotRun(resource_id))
    return ReportItem.info(
        reports.messages.ResourceRunningOnNodes(resource_id, roles_with_nodes)
    )


def ensure_resource_state(expected_running, cluster_state, resource_id):
    roles_with_nodes = _get_primitive_roles_with_nodes(
        _get_primitives_for_state_check(
            cluster_state, resource_id, expected_running
        )
    )
    if not roles_with_nodes:
        return ReportItem(
            reports.item.ReportItemSeverity(
                reports.ReportItemSeverity.INFO
                if not expected_running
                else reports.ReportItemSeverity.ERROR
            ),
            reports.messages.ResourceDoesNotRun(resource_id),
        )
    return ReportItem(
        reports.item.ReportItemSeverity(
            reports.ReportItemSeverity.INFO
            if expected_running
            else reports.ReportItemSeverity.ERROR
        ),
        reports.messages.ResourceRunningOnNodes(resource_id, roles_with_nodes),
    )


def ensure_resource_running(cluster_state, resource_id):
    return ensure_resource_state(
        expected_running=True,
        cluster_state=cluster_state,
        resource_id=resource_id,
    )


def is_resource_managed(cluster_state, resource_id):
    """
    Check if the resource is managed

    etree cluster_state -- status of the cluster
    string resource_id -- id of the resource
    """
    primitive_list = cluster_state.xpath(
        """
        .//resource[{predicate_id}]
        |
        .//group[{predicate_id}]/resource
        """.format(
            predicate_id=_id_xpath_predicate(resource_id)
        )
    )
    if primitive_list:
        for primitive in primitive_list:
            if is_false(primitive.attrib.get("managed", "")):
                return False
            parent = find_parent(primitive, ["clone", "bundle"])
            if parent is not None and is_false(
                parent.attrib.get("managed", "")
            ):
                return False
        return True

    parent_list = cluster_state.xpath(
        """
        .//clone[@id="{0}"]
        |
        .//bundle[@id="{0}"]
        """.format(
            resource_id
        )
    )
    for parent in parent_list:
        if is_false(parent.attrib.get("managed", "")):
            return False
        for primitive in parent.xpath(".//resource"):
            if is_false(primitive.attrib.get("managed", "")):
                return False
        return True

    raise ResourceNotFound(resource_id)



from dataclasses import dataclass

from .common import join_multilines


@dataclass(frozen=True)
class ExecutorResult:
    retval: int
    stdout: str
    stderr: str

    @property
    def joined_output(self) -> str:
        return join_multilines([self.stderr, self.stdout])



from unittest import mock, TestCase

from pcs_test.tools.misc import dict_to_modifiers

from pcs.common.interface import dto
from pcs.common.pacemaker.resource.relations import (
    RelationEntityDto,
    ResourceRelationDto,
    ResourceRelationType,
)
from pcs.cli.common.errors import CmdLineInputError
from pcs.cli.resource import relations


class ShowResourceRelationsCmd(TestCase):
    def setUp(self):
        self.lib_call = mock.Mock()
        self.lib = mock.Mock(spec_set=["resource"])
        self.lib.resource = mock.Mock(spec_set=["get_resource_relations_tree"])
        self.lib.resource.get_resource_relations_tree = self.lib_call
        self.lib_call.return_value = dto.to_dict(
            ResourceRelationDto(
                RelationEntityDto(
                    "d1",
                    ResourceRelationType.RSC_PRIMITIVE,
                    [],
                    {
                        "class": "ocf",
                        "provider": "pacemaker",
                        "type": "Dummy",
                    },
                ),
                [
                    ResourceRelationDto(
                        RelationEntityDto(
                            "order1",
                            ResourceRelationType.ORDER,
                            [],
                            {
                                "first-action": "start",
                                "first": "d1",
                                "then-action": "start",
                                "then": "d2",
                                "kind": "Mandatory",
                                "symmetrical": "true",
                            },
                        ),
                        [
                            ResourceRelationDto(
                                RelationEntityDto(
                                    "d2",
                                    ResourceRelationType.RSC_PRIMITIVE,
                                    [],
                                    {
                                        "class": "ocf",
                                        "provider": "heartbeat",
                                        "type": "Dummy",
                                    },
                                ),
                                [],
                                False,
                            ),
                        ],
                        False,
                    ),
                    ResourceRelationDto(
                        RelationEntityDto(
                            "inner:g1",
                            ResourceRelationType.INNER_RESOURCES,
                            [],
                            {},
                        ),
                        [
                            ResourceRelationDto(
                                RelationEntityDto(
                                    "g1", ResourceRelationType.RSC_GROUP, [], {}
                                ),
                                [],
                                True,
                            ),
                        ],
                        False,
                    ),
                ],
                False,
            )
        )

    def test_no_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            relations.show_resource_relations_cmd(
                self.lib, [], dict_to_modifiers({})
            )
        self.assertIsNone(cm.exception.message)

    def test_more_args(self):
        with self.assertRaises(CmdLineInputError) as cm:
            relations.show_resource_relations_cmd(
                self.lib, ["a1", "a2"], dict_to_modifiers({})
            )
        self.assertIsNone(cm.exception.message)

    @mock.patch("pcs.cli.resource.relations.print")
    def test_success(self, mock_print):
        relations.show_resource_relations_cmd(
            self.lib, ["d1"], dict_to_modifiers({})
        )
        self.lib_call.assert_called_once_with("d1")
        self.assertEqual(
            [
                mock.call("d1"),
                mock.call("|- inner resource(s)"),
                mock.call("|  `- g1 [displayed elsewhere]"),
                mock.call("`- order"),
                mock.call("   |  start d1 then start d2"),
                mock.call("   |  kind=Mandatory symmetrical=true"),
                mock.call("   `- d2"),
            ],
            mock_print.call_args_list,
        )

    @mock.patch("pcs.cli.resource.relations.print")
    def test_verbose(self, mock_print):
        relations.show_resource_relations_cmd(
            self.lib, ["d1"], dict_to_modifiers({"full": True})
        )
        self.lib_call.assert_called_once_with("d1")
        self.assertEqual(
            [
                mock.call("d1 (resource: ocf:pacemaker:Dummy)"),
                mock.call("|- inner resource(s) (None)"),
                mock.call("|  `- g1 (resource: group) [displayed elsewhere]"),
                mock.call("`- order (None)"),
                mock.call("   |  start d1 then start d2"),
                mock.call("   |  kind=Mandatory symmetrical=true"),
                mock.call("   `- d2 (resource: ocf:heartbeat:Dummy)"),
            ],
            mock_print.call_args_list,
        )


def _fixture_dummy(_id):
    return RelationEntityDto(
        _id,
        ResourceRelationType.RSC_PRIMITIVE,
        [],
        {
            "class": "ocf",
            "provider": "pacemaker",
            "type": "Dummy",
        },
    )


D1_PRIMITIVE = _fixture_dummy("d1")
D2_PRIMITIVE = _fixture_dummy("d2")


def _fixture_res_rel_dto(ent):
    return ResourceRelationDto(ent, [], True)


class ResourcePrintableNode(TestCase):
    def assert_member(self, member, ent):
        self.assertTrue(isinstance(member, relations.RelationPrintableNode))
        self.assertEqual(ent, member.relation_entity)
        self.assertEqual(True, member.is_leaf)
        self.assertEqual(0, len(member.members))

    def test_from_dto(self):
        inner_ent = RelationEntityDto(
            "inner:g1", ResourceRelationType.INNER_RESOURCES, [], {}
        )
        outer_ent = RelationEntityDto(
            "outer:g1", ResourceRelationType.OUTER_RESOURCE, [], {}
        )
        order_ent1 = RelationEntityDto(
            "order1", ResourceRelationType.ORDER, [], {}
        )
        order_ent2 = RelationEntityDto(
            "order2", ResourceRelationType.ORDER, [], {}
        )
        order_set_ent = RelationEntityDto(
            "order_set", ResourceRelationType.ORDER_SET, [], {}
        )

        dto_obj = ResourceRelationDto(
            D1_PRIMITIVE,
            [
                _fixture_res_rel_dto(order_set_ent),
                _fixture_res_rel_dto(order_ent2),
                _fixture_res_rel_dto(outer_ent),
                _fixture_res_rel_dto(inner_ent),
                _fixture_res_rel_dto(order_ent1),
            ],
            False,
        )
        obj = relations.ResourcePrintableNode.from_dto(dto_obj)
        self.assertEqual(D1_PRIMITIVE, obj.relation_entity)
        self.assertEqual(False, obj.is_leaf)
        expected_members = (
            inner_ent,
            outer_ent,
            order_ent1,
            order_ent2,
            order_set_ent,
        )
        self.assertEqual(len(expected_members), len(obj.members))
        for i, member in enumerate(obj.members):
            self.assert_member(member, expected_members[i])

    def test_primitive(self):
        obj = relations.ResourcePrintableNode(D1_PRIMITIVE, [], False)
        self.assertEqual(
            "d1 (resource: ocf:pacemaker:Dummy)", obj.get_title(verbose=True)
        )
        self.assertEqual([], obj.detail)

    def test_primitive_not_verbose(self):
        obj = relations.ResourcePrintableNode(D1_PRIMITIVE, [], False)
        self.assertEqual("d1", obj.get_title(verbose=False))
        self.assertEqual([], obj.detail)

    def test_primitive_without_provider_class(self):
        obj = relations.ResourcePrintableNode(
            RelationEntityDto(
                "d1",
                ResourceRelationType.RSC_PRIMITIVE,
                [],
                {
                    "type": "Dummy",
                },
            ),
            [],
            False,
        )
        self.assertEqual("d1 (resource: Dummy)", obj.get_title(verbose=True))
        self.assertEqual([], obj.detail)

    def test_primitive_without_provider(self):
        obj = relations.ResourcePrintableNode(
            RelationEntityDto(
                "d1",
                ResourceRelationType.RSC_PRIMITIVE,
                [],
                {
                    "class": "ocf",
                    "type": "Dummy",
                },
            ),
            [],
            False,
        )
        self.assertEqual(
            "d1 (resource: ocf:Dummy)", obj.get_title(verbose=True)
        )
        self.assertEqual([], obj.detail)

    def test_primitive_without_class(self):
        obj = relations.ResourcePrintableNode(
            RelationEntityDto(
                "d1",
                ResourceRelationType.RSC_PRIMITIVE,
                [],
                {
                    "provider": "pacemaker",
                    "type": "Dummy",
                },
            ),
            [],
            False,
        )
        self.assertEqual(
            "d1 (resource: pacemaker:Dummy)", obj.get_title(verbose=True)
        )
        self.assertEqual([], obj.detail)

    def test_other(self):
        obj = relations.ResourcePrintableNode(
            RelationEntityDto("an_id", "a_type", [], {}),
            [],
            False,
        )
        self.assertEqual(
            "an_id (resource: <unknown>)", obj.get_title(verbose=True)
        )
        self.assertEqual([], obj.detail)

    def test_other_not_verbose(self):
        obj = relations.ResourcePrintableNode(
            RelationEntityDto("an_id", "a_type", [], {}),
            [],
            False,
        )
        self.assertEqual("an_id", obj.get_title(verbose=False))
        self.assertEqual([], obj.detail)


class RelationPrintableNode(TestCase):
    def setUp(self):
        self.order_entity = RelationEntityDto(
            "order1",
            ResourceRelationType.ORDER,
            [],
            {
                "id": "order1",
                "first-action": "start",
                "first": "d1",
                "then-action": "start",
                "then": "d2",
            },
        )
        self.order_set_entity = RelationEntityDto(
            "order_set_id",
            ResourceRelationType.ORDER_SET,
            [],
            {
                "id": "order_set_id",
                "sets": [
                    {
                        "members": ["d1", "d2", "d3"],
                        "metadata": {},
                    },
                    {
                        "members": ["d4", "d5", "d0"],
                        "metadata": {
                            "sequential": "true",
                            "require-all": "false",
                            "score": "10",
                        },
                    },
                ],
            },
        )

    def assert_member(self, member, ent):
        self.assertTrue(isinstance(member, relations.ResourcePrintableNode))
        self.assertEqual(ent, member.relation_entity)
        self.assertEqual(True, member.is_leaf)
        self.assertEqual(0, len(member.members))

    def test_from_dto(self):
        dto_obj = ResourceRelationDto(
            self.order_entity,
            [
                ResourceRelationDto(D2_PRIMITIVE, [], True),
                ResourceRelationDto(D1_PRIMITIVE, [], True),
            ],
            False,
        )
        obj = relations.RelationPrintableNode.from_dto(dto_obj)
        self.assertEqual(self.order_entity, obj.relation_entity)
        self.assertEqual(False, obj.is_leaf)
        self.assertEqual(2, len(obj.members))
        self.assert_member(obj.members[0], D1_PRIMITIVE)
        self.assert_member(obj.members[1], D2_PRIMITIVE)

    def test_order_not_verbose(self):
        obj = relations.RelationPrintableNode(self.order_entity, [], False)
        self.assertEqual("order", obj.get_title(verbose=False))
        self.assertEqual(["start d1 then start d2"], obj.detail)

    def test_order(self):
        obj = relations.RelationPrintableNode(self.order_entity, [], False)
        self.assertEqual("order (order1)", obj.get_title(verbose=True))
        self.assertEqual(["start d1 then start d2"], obj.detail)

    def test_order_full(self):
        self.order_entity.metadata.update(
            {
                "kind": "Optional",
                "symmetrical": "true",
                "unsupported": "value",
                "score": "1000",
            }
        )
        obj = relations.RelationPrintableNode(self.order_entity, [], False)
        self.assertEqual("order (order1)", obj.get_title(verbose=True))
        self.assertEqual(
            [
                "start d1 then start d2",
                "kind=Optional score=1000 symmetrical=true",
            ],
            obj.detail,
        )

    def test_order_set_not_verbose(self):
        obj = relations.RelationPrintableNode(self.order_set_entity, [], False)
        self.assertEqual("order set", obj.get_title(verbose=False))
        self.assertEqual(
            [
                "   set d1 d2 d3",
                "   set d4 d5 d0 (require-all=false score=10 sequential=true)",
            ],
            obj.detail,
        )

    def test_order_set(self):
        obj = relations.RelationPrintableNode(self.order_set_entity, [], False)
        self.assertEqual(
            "order set (order_set_id)", obj.get_title(verbose=True)
        )
        self.assertEqual(
            [
                "   set d1 d2 d3",
                "   set d4 d5 d0 (require-all=false score=10 sequential=true)",
            ],
            obj.detail,
        )

    def test_order_set_full(self):
        self.order_set_entity.metadata.update(
            {
                "symmetrical": "true",
                "kind": "Optional",
                "require-all": "true",
                "score": "100",
                "unsupported": "value",
            }
        )
        self.order_set_entity.metadata["sets"].append(
            {
                "members": ["d9", "d8", "d6", "d7"],
                "metadata": {
                    "sequential": "true",
                    "require-all": "false",
                    "score": "10",
                    "ordering": "value",
                    "action": "start",
                    "role": "promoted",
                    "kind": "Optional",
                    "unsupported": "value",
                },
            }
        )
        obj = relations.RelationPrintableNode(self.order_set_entity, [], False)
        self.assertEqual(
            "order set (order_set_id)", obj.get_title(verbose=True)
        )
        self.assertEqual(
            [
                "kind=Optional require-all=true score=100 symmetrical=true",
                "   set d1 d2 d3",
                "   set d4 d5 d0 (require-all=false score=10 sequential=true)",
                "   set d9 d8 d6 d7 (action=start kind=Optional ordering=value "
                "require-all=false role=promoted score=10 sequential=true)",
            ],
            obj.detail,
        )

    def test_multiple_inner_resources(self):
        obj = relations.RelationPrintableNode(
            RelationEntityDto(
                "inner:g1",
                ResourceRelationType.INNER_RESOURCES,
                ["m1", "m2", "m0"],
                {"id": "g1"},
            ),
            [],
            False,
        )
        self.assertEqual("inner resource(s) (g1)", obj.get_title(verbose=True))
        self.assertEqual(["members: m1 m2 m0"], obj.detail)

    def test_inner_resources_not_verbose(self):
        obj = relations.RelationPrintableNode(
            RelationEntityDto(
                "inner:g1",
                ResourceRelationType.INNER_RESOURCES,
                ["m0"],
                {
                    "id": "g1",
                },
            ),
            [],
            False,
        )
        self.assertEqual("inner resource(s)", obj.get_title(verbose=False))
        self.assertEqual([], obj.detail)

    def test_inner_resources(self):
        obj = relations.RelationPrintableNode(
            RelationEntityDto(
                "inner:g1",
                ResourceRelationType.INNER_RESOURCES,
                ["m0"],
                {
                    "id": "g1",
                },
            ),
            [],
            False,
        )
        self.assertEqual("inner resource(s) (g1)", obj.get_title(verbose=True))
        self.assertEqual([], obj.detail)

    def test_outer_resourcenot_verbose(self):
        obj = relations.RelationPrintableNode(
            RelationEntityDto(
                "outer:g1",
                ResourceRelationType.OUTER_RESOURCE,
                [],
                {
                    "id": "g1",
                },
            ),
            [],
            False,
        )
        self.assertEqual("outer resource", obj.get_title(verbose=False))
        self.assertEqual([], obj.detail)

    def test_outer_resource(self):
        obj = relations.RelationPrintableNode(
            RelationEntityDto(
                "outer:g1",
                ResourceRelationType.OUTER_RESOURCE,
                [],
                {
                    "id": "g1",
                },
            ),
            [],
            False,
        )
        self.assertEqual("outer resource (g1)", obj.get_title(verbose=True))
        self.assertEqual([], obj.detail)

    def test_unknown_not_verbose(self):
        obj = relations.RelationPrintableNode(
            RelationEntityDto(
                "random",
                "undifined type",
                [],
                {
                    "id": "random_id",
                },
            ),
            [],
            False,
        )
        self.assertEqual("<unknown>", obj.get_title(verbose=False))
        self.assertEqual([], obj.detail)

    def test_unknown(self):
        obj = relations.RelationPrintableNode(
            RelationEntityDto(
                "random",
                "undifined type",
                [],
                {
                    "id": "random_id",
                },
            ),
            [],
            False,
        )
        self.assertEqual("<unknown> (random_id)", obj.get_title(verbose=True))
        self.assertEqual([], obj.detail)
